import tensorflow as tf
import tensorflow.keras.backend as K
import numpy as np
import scipy as sp

def binary_dice_coefficient(smoothing_factor=0.0):

    """
    Binary dice segmentation loss.

    Note:  Assumption is that y_true is *not* a one-hot representation
    of the segmentation batch.  For use with e.g., sigmoid activation.

    Arguments
    ---------

    smoothing_factor : float
        Used to smooth value during optimization

    Returns
    -------
    Loss value (negative Dice coefficient).

    """

    def binary_dice_coefficient_fixed(y_true, y_pred):

        y_true_f = K.flatten(y_true)
        y_pred_f = K.flatten(y_pred)
        intersection = K.sum(y_true_f * y_pred_f)
        return(-1.0 * (2.0 * intersection + smoothing_factor)/
           (K.sum(y_true_f) + K.sum(y_pred_f) + smoothing_factor))

    return(binary_dice_coefficient_fixed)

def multilabel_dice_coefficient(dimensionality=3, smoothing_factor=0.0):

    """
    Multi-label dice segmentation loss

    Note:  Assumption is that y_true is a one-hot representation
    of the segmentation batch.  The background (label 0) should
    be included but is not used in the calculation.  For use with
    e.g., softmax activation.

    Arguments
    ---------
    dimensionality : dimensionality
        Image dimension

    smoothing_factor : float
        Used to smooth value during optimization

    Returns
    -------
    Loss value (negative Dice coefficient).

    Example
    -------
    >>> import ants
    >>> import antspynet
    >>> import tensorflow as tf
    >>> import numpy as np
    >>>
    >>> r16 = ants.image_read(ants.get_ants_data("r16"))
    >>> r16_seg = ants.kmeans_segmentation(r16, 3)['segmentation']
    >>> r16_array = np.expand_dims(r16_seg.numpy(), axis=0)
    >>> r16_tensor = tf.convert_to_tensor(antspynet.encode_unet(r16_array, (0, 1, 2, 3)))
    >>>
    >>> r64 = ants.image_read(ants.get_ants_data("r64"))
    >>> r64_seg = ants.kmeans_segmentation(r64, 3)['segmentation']
    >>> r64_array = np.expand_dims(r64_seg.numpy(), axis=0)
    >>> r64_tensor = tf.convert_to_tensor(antspynet.encode_unet(r64_array, (0, 1, 2, 3)))
    >>>
    >>> dice_loss = antspynet.multilabel_dice_coefficient(dimensionality=2)
    >>> loss_value = dice_loss(r16_tensor, r64_tensor).numpy()
    >>> # Compare with...
    >>> ants.label_overlap_measures(r16_seg, r64_seg)
    """

    def multilabel_dice_coefficient_fixed(y_true, y_pred):
        y_dims = K.int_shape(y_pred)

        number_of_labels = y_dims[len(y_dims)-1]

        if dimensionality == 2:
            # 2-D image
            y_true_permuted = K.permute_dimensions(y_true, pattern = (3, 0, 1, 2))
            y_pred_permuted = K.permute_dimensions(y_pred, pattern = (3, 0, 1, 2))
        elif dimensionality == 3:
            # 3-D image
            y_true_permuted = K.permute_dimensions(y_true, pattern = (4, 0, 1, 2, 3))
            y_pred_permuted = K.permute_dimensions(y_pred, pattern = (4, 0, 1, 2, 3))
        else:
            raise ValueError("Specified dimensionality not implemented.")

        y_true_label = K.gather(y_true_permuted, indices = (1))
        y_pred_label = K.gather(y_pred_permuted, indices = (1))

        y_true_label_f = K.flatten(y_true_label)
        y_pred_label_f = K.flatten(y_pred_label)
        intersection = y_true_label_f * y_pred_label_f
        union = y_true_label_f + y_pred_label_f - intersection

        numerator = K.sum(intersection)
        denominator = K.sum(union)

        if number_of_labels > 2:
            for j in range(2, number_of_labels):
                y_true_label = K.gather(y_true_permuted, indices = (j))
                y_pred_label = K.gather(y_pred_permuted, indices = (j))
                y_true_label_f = K.flatten(y_true_label)
                y_pred_label_f = K.flatten(y_pred_label)

                intersection = y_true_label_f * y_pred_label_f
                union = y_true_label_f + y_pred_label_f - intersection

                numerator = numerator + K.sum(intersection)
                denominator = denominator + K.sum(union)

        unionOverlap = numerator / denominator

        return(-1.0 * (2.0 * unionOverlap + smoothing_factor) /
        (1.0 + unionOverlap + smoothing_factor))

    return(multilabel_dice_coefficient_fixed)

def peak_signal_to_noise_ratio(y_true, y_pred):
    return(-10.0 * K.log(K.mean(K.square(y_pred - y_true))) / K.log(10.0))

def pearson_correlation_coefficient(y_true, y_pred):
    N = K.sum(K.ones_like(y_true))

    sum_x = K.sum(y_true)
    sum_y = K.sum(y_pred)
    sum_x_squared = K.sum(K.square(y_true))
    sum_y_squared = K.sum(K.square(y_pred))
    sum_xy = K.sum(y_true * y_pred)

    numerator = sum_xy - (sum_x * sum_y / N)
    denominator = K.sqrt((sum_x_squared - K.square(sum_x) / N) *
      (sum_y_squared - K.square(sum_y) / N))

    coefficient = numerator / denominator

    return(coefficient)

def categorical_focal_loss(gamma=2.0, alpha=0.25):

    def categorical_focal_loss_fixed(y_true, y_pred):
        y_pred = y_pred / K.sum(y_pred, axis=-1, keepdims=True)
        y_pred = K.clip(y_pred, K.epsilon(), 1.0 - K.epsilon())
        cross_entropy = y_true * K.log(y_pred)
        loss = alpha * K.pow(1.0 - y_pred, gamma) * cross_entropy
        return(-K.sum(loss, axis=-1))

    return(categorical_focal_loss_fixed)

def weighted_categorical_crossentropy(weights):

    weights_tensor = K.variable(weights)

    def weighted_categorical_crossentropy_fixed(y_true, y_pred):
        y_pred /= K.sum(y_pred, axis=-1, keepdims=True)
        y_pred = K.clip(y_pred, K.epsilon(), 1.0 - K.epsilon())
        loss = y_true * K.log(y_pred) * weights_tensor
        loss = -K.sum(loss, axis=-1)
        return(loss)

    return(weighted_categorical_crossentropy_fixed)

def multilabel_surface_loss(dimensionality=3):

    def multilabel_surface_loss_fixed(y_true, y_pred):
        def calculate_residual_distance_map(segmentation):
            residual_distance = np.zeros_like(segmentation)

            positive_mask = segmentation.astype(np.bool)
            if positive_mask.any():
                negative_mask = ~positive_mask
                residual_distance = \
                    (sp.ndimage.distance_transform_edt(negative_mask) * negative_mask -
                    (sp.ndimage.distance_transform_edt(positive_mask) - 1) * positive_mask)

            return(residual_distance)

        def calculate_batchwise_residual_distance_maps(y_true):
            y_true_numpy = y_true.numpy()
            return(np.array([calculate_residual_distance_map(y)
                for y in y_true_numpy]).astype(np.float32))

        y_true_distance_map = tf.py_function(
            func=calculate_batchwise_residual_distance_maps,
            inp=[y_true],
            Tout=tf.float32)

        product = y_pred * y_true_distance_map
        return(K.mean(product))

    return(multilabel_surface_loss_fixed)


def maximum_mean_discrepancy(sigma=1.0):

    def maximum_mean_discrepancy_fixed(y_true, y_pred):

        x = y_true
        y = y_pred

        def compute_kernel(x, y, sigma=1.0):

            x_size = K.shape(x)[0]
            y_size = K.shape(x)[1]
            dim = K.shape(x)[1]
            x_tiled = K.tile(K.reshape(x, K.stack([x_size, 1, dim])), K.stack([1, y_size, 1]))
            y_tiled = K.tile(K.reshape(y, K.stack([1, y_size, dim])), K.stack([x_size, 1, 1]))

            denominator = 2.0 * K.square(sigma)
            kernel_value = K.exp(-K.mean(K.square(x_tiled - y_tiled) / denominator, axis=3) / K.cast( dim, 'float32'))
            return kernel_value

        x_kernel = compute_kernel(x, x, sigma)
        y_kernel = compute_kernel(y, y, sigma)
        xy_kernel = compute_kernel(x, y, sigma)

        mmd_value = K.mean(x_kernel) + K.mean(y_kernel) - 2 * K.mean(xy_kernel)
        return mmd_value

    return(maximum_mean_discrepancy_fixed)



from .get_pretrained_network import get_pretrained_network
from .get_antsxnet_data import get_antsxnet_data

from .unet_utilities import encode_unet
from .unet_utilities import decode_unet
from .denseunet_utilities import Scale

from .spatial_transformer_network_utilities import SpatialTransformer2D, SpatialTransformer3D

from .extract_image_patches import extract_image_patches
from .reconstruct_image_from_patches import reconstruct_image_from_patches

from .regression_match_image import regression_match_image

from .super_resolution_utilities import mse, mae, psnr, ssim, gmsd
from .super_resolution_utilities import apply_super_resolution_model_to_image

from .deep_embedded_clustering_utilities import DeepEmbeddedClustering
from .deep_embedded_clustering_utilities import DeepEmbeddedClusteringModel

from .mixture_density_utilities import MixtureDensityLayer
from .mixture_density_utilities import get_mixture_density_loss_function
from .mixture_density_utilities import get_mixture_density_sampling_function
from .mixture_density_utilities import get_mixture_density_mse_function
from .mixture_density_utilities import split_mixture_parameters
from .mixture_density_utilities import mixture_density_software_max
from .mixture_density_utilities import sample_from_output

from .resample_tensor_utilities import ResampleTensorLayer2D, ResampleTensorLayer3D
from .resample_tensor_utilities import ResampleTensorToTargetTensorLayer2D, ResampleTensorToTargetTensorLayer3D

from .attention_utilities import AttentionLayer2D, AttentionLayer3D

from .custom_metrics import binary_dice_coefficient
from .custom_metrics import multilabel_dice_coefficient
from .custom_metrics import peak_signal_to_noise_ratio
from .custom_metrics import pearson_correlation_coefficient
from .custom_metrics import categorical_focal_loss
from .custom_metrics import weighted_categorical_crossentropy
from .custom_metrics import multilabel_surface_loss
from .custom_metrics import maximum_mean_discrepancy

from .custom_normalization_layers import InstanceNormalization

from .custom_activation_layers import LogSoftmax

from .cropping_and_padding_utilities import crop_image_center
from .cropping_and_padding_utilities import pad_or_crop_image_to_size
from .cropping_and_padding_utilities import pad_image_by_factor

from .histogram_warp_image_intensities import histogram_warp_image_intensities
from .simulate_bias_field import simulate_bias_field

from .randomly_transform_image_data import randomly_transform_image_data

from .preprocess_image import preprocess_brain_image
from .brain_extraction import brain_extraction
from .cortical_thickness import cortical_thickness
from .cortical_thickness import longitudinal_cortical_thickness
from .lung_extraction import lung_extraction
from .white_matter_hyperintensity_segmentation import sysu_media_wmh_segmentation
from .white_matter_hyperintensity_segmentation import ew_david
from .claustrum_segmentation import claustrum_segmentation
from .hypothalamus_segmentation import hypothalamus_segmentation
from .hippmapp3r_segmentation import hippmapp3r_segmentation
from .deep_flash import deep_flash
from .deep_atropos import deep_atropos
from .desikan_killiany_tourville_labeling import desikan_killiany_tourville_labeling
from .brain_age import brain_age
from .mri_super_resolution import mri_super_resolution
from .quality_assessment import tid_neural_image_assessment
from .lung_segmentation import el_bicho

from .neural_style_transfer import neural_style_transfer



import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import (Input, Concatenate, Dense, Activation,
                                     BatchNormalization, Reshape, Dropout,
                                     Flatten, LeakyReLU, Conv2D, Conv3D)
from tensorflow.keras.optimizers import RMSprop

from . import (create_convolutional_autoencoder_model_2d,
               create_convolutional_autoencoder_model_3d)

from ..utilities import ResampleTensorLayer2D, ResampleTensorLayer3D

import numpy as np
import os

import ants

class WassersteinGanModel(object):
    """
    Wasserstein GAN model

    Wasserstein generative adverserial network from the paper:

      https://arxiv.org/abs/1701.07875

    and ported from the Keras implementation:

      https://github.com/eriklindernoren/Keras-GAN/blob/master/srgan/srgan.py

    Arguments
    ---------
    input_image_size : tuple
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    latent_dimension : integer
        Default = 100.

    number_of_critic_iterations : integer
        Default = 5.

    clip_value : float
        Default = 0.01.

    Returns
    -------
    Keras model
        A Keras model defining the network.
    """

    def __init__(self, input_image_size, latent_dimension=100,
                 number_of_critic_iterations=5, clip_value=0.01):
        super(WassersteinGanModel, self).__init__()

        self.input_image_size = input_image_size
        self.latent_dimension = latent_dimension
        self.number_of_critic_iterations = number_of_critic_iterations
        self.clip_value = clip_value

        self.dimensionality = None
        if len(self.input_image_size) == 3:
            self.dimensionality = 2
        elif len(self.input_image_size) == 4:
            self.dimensionality = 3
        else:
            raise ValueError("Incorrect size for input_image_size.")

        optimizer = RMSprop(lr=0.00005)

        self.critic = self.build_critic()
        self.critic.compile(loss=self.wasserstein_loss, optimizer=optimizer,
                            metrics=['acc'])
        self.critic.trainable = False

        self.generator = self.build_generator()

        z = Input(shape=(self.latent_dimension,))
        image = self.generator(z)

        validity = self.critic(image)

        self.combined_model = Model(inputs=z, outputs=validity)
        self.combined_model.compile(loss=self.wasserstein_loss,
                                    optimizer=optimizer, metrics=['acc'])

    def wasserstein_loss(self, y_true, y_pred):
       return(K.mean(y_true * y_pred))

    def build_generator(self, number_of_filters_per_layer=(128, 64), kernel_size=4):

        model = Sequential()

        # To build the generator, we create the reverse encoder model
        # and simply build the reverse model

        encoder = None
        if self.dimensionality == 2:
             autoencoder, encoder = create_convolutional_autoencoder_model_2d(
                          input_image_size=self.input_image_size,
                          number_of_filters_per_layer=(*(number_of_filters_per_layer[::-1]), self.latent_dimension),
                          convolution_kernel_size=(5, 5),
                          deconvolution_kernel_size=(5, 5))
        else:
             autoencoder, encoder = create_convolutional_autoencoder_model_3d(
                          input_image_size=self.input_image_size,
                          number_of_filters_per_layer=(*(number_of_filters_per_layer[::-1]), self.latent_dimension),
                          convolution_kernel_size=(5, 5, 5),
                          deconvolution_kernel_size=(5, 5, 5))

        encoder_layers = encoder.layers

        penultimate_layer = encoder_layers[len(encoder_layers) - 2]

        model.add(Dense(units=penultimate_layer.output_shape[1],
                        input_dim=self.latent_dimension,
                        activation="relu"))

        conv_layer = encoder_layers[len(encoder_layers) - 3]
        resampled_size = conv_layer.output_shape[1:(self.dimensionality + 2)]
        model.add(Reshape(resampled_size))

        count = 0
        for i in range(len(encoder_layers) - 3, 1, -1):
            conv_layer = encoder_layers[i]
            resampled_size = conv_layer.output_shape[1:(self.dimensionality + 1)]

            if self.dimensionality == 2:
                model.add(ResampleTensorLayer2D(shape=resampled_size,
                                                interpolation_type='linear'))
                model.add(Conv2D(filters=number_of_filters_per_layer[count],
                                 kernel_size=kernel_size,
                                 padding='same'))
            else:
                model.add(ResampleTensorLayer3D(shape=resampled_size,
                                                interpolation_type='linear'))
                model.add(Conv3D(filters=number_of_filters_per_layer[count],
                                 kernel_size=kernel_size,
                                 padding='same'))
            model.add(BatchNormalization(momentum=0.8))
            model.add(Activation(activation='relu'))
            count += 1

        number_of_channels = self.input_image_size[-1]
        spatial_dimensions = self.input_image_size[:self.dimensionality]

        if self.dimensionality == 2:
            model.add(ResampleTensorLayer2D(shape=spatial_dimensions,
                                            interpolation_type='linear'))
            model.add(Conv2D(filters=number_of_channels,
                             kernel_size=kernel_size,
                             padding='same'))
        else:
            model.add(ResampleTensorLayer3D(shape=spatial_dimensions,
                                            interpolation_type='linear'))
            model.add(Conv3D(filters=number_of_channels,
                             kernel_size=kernel_size,
                             padding='same'))

        model.add(Activation(activation="tanh"))

        noise = Input(shape=(self.latent_dimension,))
        image = model(noise)

        generator = Model(inputs=noise, outputs=image)
        return(generator)

    def build_critic(self, number_of_filters_per_layer=(16, 32, 64, 128),
                            kernel_size=3, dropout_rate=0.25):
        model = Sequential()

        for i in range(len(number_of_filters_per_layer)):

            strides = 2
            if i == len(number_of_filters_per_layer) - 1:
                strides=1

            if self.dimensionality == 2:
                model.add(Conv2D(input_shape=self.input_image_size,
                                 filters=number_of_filters_per_layer[i],
                                 kernel_size = kernel_size,
                                 strides = strides,
                                 padding='same'))
            else:
                model.add(Conv3D(input_shape=self.input_image_size,
                                 filters=number_of_filters_per_layer[i],
                                 kernel_size = kernel_size,
                                 strides = strides,
                                 padding='same'))

            if i > 0:
                model.add(BatchNormalization(momentum=0.8))

            model.add(LeakyReLU(alpha=0.2))
            model.add(Dropout(rate=dropout_rate))

        model.add(Flatten())
        model.add(Dense(units=1))

        image = Input(shape=self.input_image_size)

        validity = model(image)

        critic = Model(inputs=image, outputs=validity)

        return(critic)

    def train(self, X_train, number_of_epochs, batch_size=128,
              sample_interval=None, sample_file_prefix='sample'):
        valid = -np.ones((batch_size, 1))
        fake = np.ones((batch_size, 1))

        for epoch in range(number_of_epochs):

            # train critic

            for c in range(self.number_of_critic_iterations):
                indices = np.random.randint(0, X_train.shape[0] - 1, batch_size)
                X_valid_batch = X_train[indices]

                noise = np.random.normal(0, 1, (batch_size, self.latent_dimension))
                X_fake_batch = self.generator.predict(noise)

                d_loss_real = self.critic.train_on_batch(X_valid_batch, valid)
                d_loss_fake = self.critic.train_on_batch(X_fake_batch, fake)
                d_loss = 0.5 * np.add(d_loss_real, d_loss_fake)

                # clip critic weights

                for i in range(len(self.critic.layers)):
                    weights = self.critic.layers[i].get_weights()
                    for j in range(len(weights)):
                       weights[j] = np.clip(weights[j], -self.clip_value, self.clip_value)
                    self.critic.layers[i].set_weights(weights)

            # train generator

            noise = np.random.normal(0, 1, (batch_size, self.latent_dimension))
            g_loss = self.combined_model.train_on_batch(noise, valid)

            print("Epoch ", epoch, ": [Critic loss: ", 1.0 - d_loss[0],
                  "] ", "[Generator loss: ", 1.0 - g_loss[0])

            if self.dimensionality == 2:
                if sample_interval != None:
                    if epoch % sample_interval == 0:

                        # Do a 5x5 grid

                        predicted_batch_size = 5 * 5
                        noise = np.random.normal(0, 1, (predicted_batch_size, self.latent_dimension))
                        X_generated = self.generator.predict(noise)

                        # Convert to [0,255] to write as jpg using ANTsPy

                        X_generated = (255 * (X_generated - X_generated.min()) /
                          (X_generated.max() - X_generated.min()))
                        X_generated = np.squeeze(X_generated)
                        X_generated = np.uint8(X_generated)

                        X_tiled = np.zeros((5 * X_generated.shape[1], 5 * X_generated.shape[2]), dtype=np.uint8)
                        for i in range(5):
                            indices_i = (i * X_generated.shape[1], (i + 1) * X_generated.shape[1])
                            for j in range(5):
                                indices_j = (j * X_generated.shape[2], (j + 1) * X_generated.shape[2])
                                X_tiled[indices_i[0]:indices_i[1], indices_j[0]:indices_j[1]] = \
                                  np.squeeze(X_generated[i * 5 + j, :, :])

                        X_generated_image = ants.from_numpy(np.transpose(X_tiled))

                        image_file_name = sample_file_prefix + "_iteration" + str(epoch) + ".jpg"
                        dir_name = os.path.dirname(sample_file_prefix)
                        if not os.path.exists(dir_name):
                            os.mkdir(dir_name)

                        print("   --> writing sample image: ", image_file_name)
                        ants.image_write(X_generated_image, image_file_name)








from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Average, Add,
                          Conv2D, Conv2DTranspose,
                          Conv3D, Conv3DTranspose)

def create_denoising_auto_encoder_super_resolution_model_2d(input_image_size,
                                                            convolution_kernel_sizes=[(3, 3), (5, 5)],
                                                            number_of_encoding_layers=2,
                                                            number_of_filters=64
                                                           ):
    """
    2-D implementation of the denoising autoencoder image super resolution deep learning architecture.

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    convolution_kernel_sizes : list of 2-d tuples
        specifies the kernel size at each convolution layer.  Default values are
        the same as given in the original paper.  The length of kernel size list
        must be 1 greater than the tuple length of the number of filters.

    number_of_encoding_layers : integer
       The number of encoding layers.

    number_of_filters : integer
       The number of filters for each encoding layer.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_denoising_auto_encoder_super_resolution_model_2d((128, 128, 1))
    >>> model.summary()
    """

    inputs = Input(shape = input_image_size)

    outputs = inputs

    encoding_convolution_layers = []
    for i in range(number_of_encoding_layers):
        if i == 0:
            outputs = Conv2D(filters=number_of_filters,
                             kernel_size=convolution_kernel_sizes[0],
                             activation='relu',
                             padding='same')(outputs)
        else:
            layer = Conv2D(filters=number_of_filters,
                           kernel_size=convolution_kernel_sizes[0],
                           activation='relu',
                           padding='same')(outputs)
            encoding_convolution_layers.append(layer)

    outputs = encoding_convolution_layers[-1]

    for i in range(number_of_encoding_layers):
        index = len(encoding_convolution_layers) - i - 1
        deconvolution = Conv2DTranspose(filters=number_of_filters,
                                        kernel_size=convolution_kernel_sizes[0],
                                        padding='same',
                                        activation='relu')(outputs)
        outputs = Add()([encoding_convolution_layers[index], deconvolution])

    number_of_channels = input_image_size[-1]

    outputs = Conv2D(filters=number_of_channels,
                     kernel_size=convolution_kernel_sizes[1],
                     activation='linear',
                     padding='same')(outputs)

    sr_model = Model(inputs=inputs, outputs=outputs)

    return(sr_model)

def create_denoising_auto_encoder_super_resolution_model_3d(input_image_size,
                                                            convolution_kernel_sizes=[(3, 3, 3), (5, 5, 5)],
                                                            number_of_encoding_layers=2,
                                                            number_of_filters=64
                                                           ):
    """
    2-D implementation of the denoising autoencoder image super resolution deep learning architecture.

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    convolution_kernel_sizes : list of 3-d tuples
        specifies the kernel size at each convolution layer.  Default values are
        the same as given in the original paper.  The length of kernel size list
        must be 1 greater than the tuple length of the number of filters.

    number_of_encoding_layers : integer
       The number of encoding layers.

    number_of_filters : integer
       The number of filters for each encoding layer.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_denoising_auto_encoder_super_resolution_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    inputs = Input(shape = input_image_size)

    outputs = inputs

    encoding_convolution_layers = []
    for i in range(number_of_encoding_layers):
        if i == 0:
            outputs = Conv3D(filters=number_of_filters,
                             kernel_size=convolution_kernel_sizes[0],
                             activation='relu',
                             padding='same')(outputs)
        else:
            layer = Conv3D(filters=number_of_filters,
                           kernel_size=convolution_kernel_sizes[0],
                           activation='relu',
                           padding='same')(outputs)
            encoding_convolution_layers.append(layer)

    outputs = encoding_convolution_layers[-1]

    for i in range(number_of_encoding_layers):
        index = len(encoding_convolution_layers) - i - 1
        deconvolution = Conv3DTranspose(filters=number_of_filters,
                                        kernel_size=convolution_kernel_sizes[0],
                                        padding='same',
                                        activation='relu')(outputs)
        outputs = Add()([encoding_convolution_layers[index], deconvolution])

    number_of_channels = input_image_size[-1]

    outputs = Conv3D(filters=number_of_channels,
                     kernel_size=convolution_kernel_sizes[1],
                     activation='linear',
                     padding='same')(outputs)

    sr_model = Model(inputs=inputs, outputs=outputs)

    return(sr_model)




import numpy as np
import tensorflow as tf
import ants

def preprocess_brain_image(image,
                           truncate_intensity=(0.01, 0.99),
                           brain_extraction_modality=None,
                           template_transform_type=None,
                           template="biobank",
                           do_bias_correction=True,
                           return_bias_field=False,
                           do_denoising=True,
                           intensity_matching_type=None,
                           reference_image=None,
                           intensity_normalization_type=None,
                           antsxnet_cache_directory=None,
                           verbose=True):

    """
    Basic preprocessing pipeline for T1-weighted brain MRI

    Standard preprocessing steps that have been previously described
    in various papers including the cortical thickness pipeline:

         https://www.ncbi.nlm.nih.gov/pubmed/24879923

    Arguments
    ---------
    image : ANTsImage
        input image

    truncate_intensity : 2-length tuple
        Defines the quantile threshold for truncating the image intensity

    brain_extraction_modality : string or None
        Perform brain extraction using antspynet tools.  One of "t1", "t1v0",
        "t1nobrainer", "t1combined", "flair", "t2", "bold", "fa", "t1infant",
        "t2infant", or None.

    template_transform_type : string
        See details in help for ants.registration.  Typically "Rigid" or
        "Affine".

    template : ANTs image (not skull-stripped)
        Alternatively, one can specify the default "biobank" or "croppedMni152"
        to download and use premade templates.

    do_bias_correction : boolean
        Perform N4 bias field correction.

    return_bias_field : boolean
        If True, return bias field as an additional output *without* bias
        correcting the preprocessed image.

    do_denoising : boolean
        Perform non-local means denoising.

    intensity_matching_type : string
        Either "regression" or "histogram". Only is performed if reference_image
        is not None.

    reference_image : ANTs image
        Reference image for intensity matching.

    intensity_normalization_type : string
        Either rescale the intensities to [0,1] (i.e., "01") or zero-mean, unit variance
        (i.e., "0mean").  If None normalization is not performed.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    Dictionary with preprocessing information ANTs image (i.e., source_image) matched to the
    (reference_image).

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> preprocessed_image = preprocess_brain_image(image, do_brain_extraction=False)
    """

    from ..utilities import brain_extraction
    from ..utilities import regression_match_image
    from ..utilities import get_antsxnet_data

    preprocessed_image = ants.image_clone(image)

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    # Truncate intensity
    if truncate_intensity is not None:
        quantiles = (image.quantile(truncate_intensity[0]), image.quantile(truncate_intensity[1]))
        if verbose == True:
            print("Preprocessing:  truncate intensities ( low =", quantiles[0], ", high =", quantiles[1], ").")

        preprocessed_image[image < quantiles[0]] = quantiles[0]
        preprocessed_image[image > quantiles[1]] = quantiles[1]

    # Brain extraction
    mask = None
    if brain_extraction_modality is not None:
        if verbose == True:
            print("Preprocessing:  brain extraction.")

        probability_mask = brain_extraction(preprocessed_image, modality=brain_extraction_modality,
            antsxnet_cache_directory=antsxnet_cache_directory, verbose=verbose)
        mask = ants.threshold_image(probability_mask, 0.5, 1, 1, 0)
        mask = ants.morphology(mask,"close",6).iMath_fill_holes()

    # Template normalization
    transforms = None
    if template_transform_type is not None:
        template_image = None
        if isinstance(template, str):
            template_file_name_path = get_antsxnet_data(template, antsxnet_cache_directory=antsxnet_cache_directory)
            template_image = ants.image_read(template_file_name_path)
        else:
            template_image = template

        if mask is None:
            registration = ants.registration(fixed=template_image, moving=preprocessed_image,
                type_of_transform=template_transform_type, verbose=verbose)
            preprocessed_image = registration['warpedmovout']
            transforms = dict(fwdtransforms=registration['fwdtransforms'],
                              invtransforms=registration['invtransforms'])
        else:
            template_probability_mask = brain_extraction(template_image, antsxnet_cache_directory=antsxnet_cache_directory, verbose=verbose)
            template_mask = ants.threshold_image(template_probability_mask, 0.5, 1, 1, 0)
            template_brain_image = template_mask * template_image

            preprocessed_brain_image = preprocessed_image * mask

            registration = ants.registration(fixed=template_brain_image, moving=preprocessed_brain_image,
                type_of_transform=template_transform_type, verbose=verbose)
            transforms = dict(fwdtransforms=registration['fwdtransforms'],
                              invtransforms=registration['invtransforms'])

            preprocessed_image = ants.apply_transforms(fixed = template_image, moving = preprocessed_image,
                transformlist=registration['fwdtransforms'], interpolator="linear", verbose=verbose)
            mask = ants.apply_transforms(fixed = template_image, moving = mask,
                transformlist=registration['fwdtransforms'], interpolator="genericLabel", verbose=verbose)

    # Do bias correction
    bias_field = None
    if do_bias_correction == True:
        if verbose == True:
            print("Preprocessing:  brain correction.")
        n4_output = None
        if mask is None:
            n4_output = ants.n4_bias_field_correction(preprocessed_image, shrink_factor=4, return_bias_field=return_bias_field, verbose=verbose)
        else:
            n4_output = ants.n4_bias_field_correction(preprocessed_image, mask, shrink_factor=4, return_bias_field=return_bias_field, verbose=verbose)
        if return_bias_field == True:
            bias_field = n4_output
        else:
            preprocessed_image = n4_output

    # Denoising
    if do_denoising == True:
        if verbose == True:
            print("Preprocessing:  denoising.")

        if mask is None:
            preprocessed_image = ants.denoise_image(preprocessed_image, shrink_factor=1)
        else:
            preprocessed_image = ants.denoise_image(preprocessed_image, mask, shrink_factor=1)

    # Image matching
    if reference_image is not None and intensity_matching_type is not None:
        if verbose == True:
            print("Preprocessing:  intensity matching.")

        if intensity_matching_type == "regression":
            preprocessed_image = regression_match_image(preprocessed_image, reference_image)
        elif intensity_matching_type == "histogram":
            preprocessed_image = ants.histogram_match_image(preprocessed_image, reference_image)
        else:
            raise ValueError("Unrecognized intensity_matching_type.")

    # Intensity normalization
    if intensity_normalization_type is not None:
        if verbose == True:
            print("Preprocessing:  intensity normalization.")

        if intensity_normalization_type == "01":
            preprocessed_image = (preprocessed_image - preprocessed_image.min())/(preprocessed_image.max() - preprocessed_image.min())
        elif intensity_normalization_type == "0mean":
            preprocessed_image = (preprocessed_image - preprocessed_image.mean())/preprocessed_image.std()
        else:
            raise ValueError("Unrecognized intensity_normalization_type.")

    return_dict = {'preprocessed_image' : preprocessed_image}
    if mask is not None:
        return_dict['brain_mask'] = mask
    if bias_field is not None:
        return_dict['bias_field'] = bias_field
    if transforms is not None:
        return_dict['template_transforms'] = transforms

    return(return_dict)



from setuptools import setup

setup(name='antspynet',
      version='0.1.1',
      description='A collection of deep learning architectures ported to the python language and tools for basic medical image processing.',
      url='http://github.com/ntustison/antspynet',
      author='Nicholas J. Tustison and Brian B. Avants and Nick Cullen',
      author_email='ntustison@gmail.com',
      packages=['antspynet','antspynet/architectures','antspynet/utilities'],
      install_requires=['antspyx','keras','scikit-learn','tensorflow','tensorflow-probability','numpy','requests','statsmodels','matplotlib'],
      zip_safe=False)




import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Dropout, BatchNormalization, Add,
                                     ThresholdedReLU, Concatenate, Dense,
                                     Conv2D, Conv2DTranspose,
                                     MaxPooling2D, UpSampling2D,
                                     Conv3D, Conv3DTranspose,
                                     MaxPooling3D, UpSampling3D)

from tensorflow.keras import initializers
from tensorflow.keras import regularizers

def create_resunet_model_2d(input_image_size,
                            number_of_outputs=1,
                            number_of_filters_at_base_layer=32,
                            bottle_neck_block_depth_schedule=(3, 4),
                            convolution_kernel_size=(3, 3),
                            deconvolution_kernel_size=(2, 2),
                            dropout_rate=0.0,
                            weight_decay=0.0,
                            mode='classification'
                           ):
    """
    2-D implementation of the Resnet + U-net deep learning architecture.

    Creates a keras model of the U-net + ResNet deep learning architecture for
    image segmentation and regression with the paper available here:

            https://arxiv.org/abs/1608.04117

    This particular implementation was ported from the following python
    implementation:

            https://github.com/veugene/fcn_maker/


    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The
        shape (or dimension) of that tensor is the image dimensions followed by
        the number of channels (e.g., red, green, and blue).  The batch size
        (i.e., number of training images) is not specified a priori.

    number_of_outputs : integer
        Meaning depends on the mode.  For 'classification' this is the number of
        segmentation labels.  For 'regression' this is the number of outputs.

    number_of_filters_at_base_layer : integer
        Number of filters at the beginning and end of the 'U'.  Doubles at each
        descending/ascending layer.

    bottle_neck_block_depth_schedule : tuple
        Tuple that provides the encoding layer schedule for the number of bottleneck
        blocks per long skip connection.

    convolution_kernel_size : tuple of length 2
        2-d vector defining the kernel size during the encoding path

    deconvolution_kernel_size : tuple of length 2
        2-d vector defining the kernel size during the decoding

    dropout_rate : scalar
        Float between 0 and 1 to use between dense layers.

    weight_decay : scalar
        Weighting parameter for L2 regularization of the kernel weights of the
        convolution layers.  Default = 0.0.

    mode : string
        'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_resunet_model_2d((128, 128, 1))
    >>> model.summary()
    """

    def simple_block_2d(input, number_of_filters,
                        downsample=False, upsample=False,
                        convolution_kernel_size=(3, 3),
                        deconvolution_kernel_size=(2, 2),
                        weight_decay=0.0, dropout_rate=0.0):

        number_of_output_filters = number_of_filters

        output = BatchNormalization()(input)
        output = ThresholdedReLU(theta = 0)(output)

        if downsample:
            output = MaxPooling2D(pool_size=(2, 2))(output)

        output = Conv2D(filters=number_of_filters,
                        kernel_size=convolution_kernel_size,
                        padding='same',
                        kernel_regularizer=regularizers.l2(weight_decay))(output)

        if upsample:
            output = Conv2DTranspose(filters=number_of_filters,
                                     kernel_size=deconvolution_kernel_size,
                                     padding='same',
                                     kernel_initializer=initializers.he_normal(),
                                     kernel_regularizer=regularizers.l2(weight_decay))(output)
            output = UpSampling2D(size=(2, 2))(output)

        if dropout_rate > 0.0:
           output=Dropout(rate=dropout_rate)(output)

        # Modify the input so that it has the same size as the output

        if downsample:
            input = Conv2D(filters=number_of_output_filters,
                            kernel_size=(1, 1),
                            strides=(2, 2),
                            padding='same')(input)
        elif upsample:
            input = Conv2DTranspose(filters=number_of_output_filters,
                                    kernel_size=(1, 1),
                                    padding='same')(input)
            input = UpSampling2D(size=(2, 2))(input)
        elif number_of_filters != number_of_output_filters:
            input = Conv2D(filters=number_of_output_filters,
                            kernel_size=(1, 1),
                            padding='same')(input)

        output = skip_connection(input, output)

        return(output)

    def bottle_neck_block_2d(input, number_of_filters, downsample=False,
                             upsample=False, deconvolution_kernel_size=(2, 2),
                             weight_decay=0.0, dropout_rate=0.0):

        output = input

        number_of_output_filters = number_of_filters

        if downsample:
            output = BatchNormalization()(output)
            output = ThresholdedReLU(theta = 0)(output)

            output = Conv2D(filters=number_of_filters,
                            kernel_size=(1, 1),
                            strides=(2, 2),
                            kernel_initializer=initializers.he_normal(),
                            kernel_regularizer=regularizers.l2(weight_decay))(output)

        output = BatchNormalization()(output)
        output = ThresholdedReLU(theta = 0)(output)

        output = Conv2D(filters=number_of_filters,
                        kernel_size=(1, 1),
                        kernel_initializer=initializers.he_normal(),
                        kernel_regularizer=regularizers.l2(weight_decay))(output)

        output = BatchNormalization()(output)
        output = ThresholdedReLU(theta = 0)(output)

        if upsample:
            output = Conv2DTranspose(filters=number_of_filters,
                                     kernel_size=deconvolution_kernel_size,
                                     padding='same',
                                     kernel_initializer=initializers.he_normal(),
                                     kernel_regularizer=regularizers.l2(weight_decay))(output)
            output = UpSampling2D(size=(2, 2))(output)

        output = Conv2D(filters=(number_of_filters * 4),
                        kernel_size=(1, 1),
                        kernel_initializer=initializers.he_normal(),
                        kernel_regularizer=regularizers.l2(weight_decay))(output)

        number_of_output_filters = number_of_filters * 4

        if dropout_rate > 0.0:
           output=Dropout(rate=dropout_rate)(output)

        # Modify the input so that it has the same size as the output

        if downsample:
            input = Conv2D(filters=number_of_output_filters,
                            kernel_size=(1, 1),
                            strides=(2, 2),
                            padding='same')(input)
        elif upsample:
            input = Conv2DTranspose(filters=number_of_output_filters,
                                    kernel_size=(1, 1),
                                    padding='same')(input)
            input = UpSampling2D(size=(2, 2))(input)
        elif number_of_filters != number_of_output_filters:
            input = Conv2D(filters=number_of_output_filters,
                            kernel_size=(1, 1),
                            padding='valid')(input)

        output = skip_connection(input, output)

        return(output)

    def skip_connection(source, target, merge_mode='sum'):
        layer_list = [source, target]

        output = None
        if merge_mode == 'sum':
            output = Add()(layer_list)
        else:
            channel_axis = 0
            if K.image_data_format() == 'channels_last':
                channel_axis = -1
            output = Concatenate(axis=channel_axis)(layer_list)

        return(output)

    inputs = Input(shape = input_image_size)

    encoding_layers_with_long_skip_connections = []
    encoding_layer_count = 1

    # Preprocessing layer

    model = Conv2D(filters=number_of_filters_at_base_layer,
                   kernel_size=convolution_kernel_size,
                   activation='relu',
                   padding='same',
                   kernel_initializer=initializers.he_normal(),
                   kernel_regularizer=regularizers.l2(weight_decay))(inputs)

    encoding_layers_with_long_skip_connections.append(model)
    encoding_layer_count += 1

    # Encoding initialization path

    model = simple_block_2d(model, number_of_filters_at_base_layer, downsample=True,
                            convolution_kernel_size=convolution_kernel_size,
                            deconvolution_kernel_size=deconvolution_kernel_size,
                            weight_decay=weight_decay, dropout_rate=dropout_rate)

    encoding_layers_with_long_skip_connections.append(model)
    encoding_layer_count += 1

    # Encoding main path

    number_of_bottle_neck_layers = len(bottle_neck_block_depth_schedule)
    for i in range(number_of_bottle_neck_layers):
        number_of_filters = number_of_filters_at_base_layer * 2**i

        for j in range(bottle_neck_block_depth_schedule[i]):

            do_downsample = False
            if j == 0:
                do_downsample = True
            else:
                do_downsample = False

            model = bottle_neck_block_2d(model, number_of_filters=number_of_filters,
                                         downsample=do_downsample,
                                         deconvolution_kernel_size=deconvolution_kernel_size,
                                         weight_decay=weight_decay, dropout_rate=dropout_rate)

            if j == (bottle_neck_block_depth_schedule[i] - 1):
               encoding_layers_with_long_skip_connections.append(model)
               encoding_layer_count += 1

    encoding_layer_count -= 1

    # Transition path

    number_of_filters = number_of_filters_at_base_layer * 2**number_of_bottle_neck_layers

    model = bottle_neck_block_2d(model, number_of_filters=number_of_filters,
                                 downsample=True,
                                 deconvolution_kernel_size=deconvolution_kernel_size,
                                 weight_decay=weight_decay, dropout_rate=dropout_rate)
    model = bottle_neck_block_2d(model, number_of_filters=number_of_filters,
                                 upsample=True,
                                 deconvolution_kernel_size=deconvolution_kernel_size,
                                 weight_decay=weight_decay, dropout_rate=dropout_rate)

    # Decoding main path

    number_of_bottle_neck_layers = len(bottle_neck_block_depth_schedule)
    for i in range(number_of_bottle_neck_layers):
        number_of_filters = (number_of_filters_at_base_layer *
                             2**(number_of_bottle_neck_layers - i - 1))

        for j in range(bottle_neck_block_depth_schedule[number_of_bottle_neck_layers - i - 1]):

            do_upsample = False
            if j == bottle_neck_block_depth_schedule[number_of_bottle_neck_layers - i - 1] - 1:
                do_upsample = True
            else:
                do_upsample = False

            model = bottle_neck_block_2d(model, number_of_filters=number_of_filters,
                                         upsample=do_upsample,
                                         deconvolution_kernel_size=deconvolution_kernel_size,
                                         weight_decay=weight_decay, dropout_rate=dropout_rate)

            if j == 0:
               model = Conv2D(filters=(number_of_filters * 4),
                              kernel_size=(1, 1),
                              padding='same')(model)
               model = skip_connection(encoding_layers_with_long_skip_connections[encoding_layer_count - 1], model)
               encoding_layer_count -= 1

    # Decoding initialization path

    model = simple_block_2d(model, number_of_filters_at_base_layer, upsample=True,
                            convolution_kernel_size=convolution_kernel_size,
                            deconvolution_kernel_size=deconvolution_kernel_size,
                            weight_decay=weight_decay, dropout_rate=dropout_rate)

    # Postprocessing layer

    model = Conv2D(filters=number_of_filters_at_base_layer,
                   kernel_size=convolution_kernel_size,
                   activation='relu',
                   padding='same',
                   kernel_initializer=initializers.he_normal(),
                   kernel_regularizer=regularizers.l2(weight_decay))(model)
    encoding_layer_count -= 1

    model = skip_connection(encoding_layers_with_long_skip_connections[encoding_layer_count - 1], model)

    model = BatchNormalization()(model)
    model = ThresholdedReLU(theta = 0)(model)

    convActivation = ''

    if mode == 'classification':
        convActivation = 'softmax'
    elif mode == 'regression':
        convActivation = 'linear'
    else:
        raise ValueError('mode must be either `classification` or `regression`.')

    outputs = Conv2D(filters=number_of_outputs,
                     kernel_size=(1, 1),
                     activation = convActivation,
                     kernel_regularizer=regularizers.l2(weight_decay))(model)

    resunet_model = Model(inputs=inputs, outputs=outputs)

    return resunet_model

def create_resunet_model_3d(input_image_size,
                            number_of_outputs=1,
                            number_of_filters_at_base_layer=32,
                            bottle_neck_block_depth_schedule=(3, 4),
                            convolution_kernel_size=(3, 3, 3),
                            deconvolution_kernel_size=(2, 2, 2),
                            dropout_rate=0.0,
                            weight_decay=0.0,
                            mode='classification'
                           ):
    """
    3-D implementation of the Resnet + U-net deep learning architecture.

    Creates a keras model of the U-net + ResNet deep learning architecture for
    image segmentation and regression with the paper available here:

            https://arxiv.org/abs/1608.04117

    This particular implementation was ported from the following python
    implementation:

            https://github.com/veugene/fcn_maker/


    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The
        shape (or dimension) of that tensor is the image dimensions followed by
        the number of channels (e.g., red, green, and blue).  The batch size
        (i.e., number of training images) is not specified a priori.

    number_of_outputs : integer
        Meaning depends on the mode.  For 'classification' this is the number of
        segmentation labels.  For 'regression' this is the number of outputs.

    number_of_filters_at_base_layer : integer
        Number of filters at the beginning and end of the 'U'.  Doubles at each
        descending/ascending layer.

    bottle_neck_block_depth_schedule : tuple
        Tuple that provides the encoding layer schedule for the number of bottleneck
        blocks per long skip connection.

    convolution_kernel_size : tuple of length 3
        3-d vector defining the kernel size during the encoding path

    deconvolution_kernel_size : tuple of length 3
        3-d vector defining the kernel size during the decoding

    dropout_rate : scalar
        Float between 0 and 1 to use between dense layers.

    weight_decay : scalar
        Weighting parameter for L2 regularization of the kernel weights of the
        convolution layers.  Default = 0.0.

    mode : string
        'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_resunet_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    def simple_block_3d(input, number_of_filters,
                        downsample=False, upsample=False,
                        convolution_kernel_size=(3, 3, 3),
                        deconvolution_kernel_size=(2, 2, 2),
                        weight_decay=0.0, dropout_rate=0.0):

        number_of_output_filters = number_of_filters

        output = BatchNormalization()(input)
        output = ThresholdedReLU(theta = 0)(output)

        if downsample:
            output = MaxPooling3D(pool_size=(2, 2, 2))(output)

        output = Conv3D(filters=number_of_filters,
                        kernel_size=convolution_kernel_size,
                        padding='same',
                        kernel_regularizer=regularizers.l2(weight_decay))(output)

        if upsample:
            output = Conv3DTranspose(filters=number_of_filters,
                                     kernel_size=deconvolution_kernel_size,
                                     padding='same',
                                     kernel_initializer=initializers.he_normal(),
                                     kernel_regularizer=regularizers.l2(weight_decay))(output)
            output = UpSampling3D(size=(2, 2, 2))(output)

        if dropout_rate > 0.0:
           output=Dropout(rate=dropout_rate)(output)

        # Modify the input so that it has the same size as the output

        if downsample:
            input = Conv3D(filters=number_of_output_filters,
                            kernel_size=(1, 1, 1),
                            strides=(2, 2, 2),
                            padding='same')(input)
        elif upsample:
            input = Conv3DTranspose(filters=number_of_output_filters,
                                    kernel_size=(1, 1, 1),
                                    padding='same')(input)
            input = UpSampling3D(size=(2, 2, 2))(input)
        elif number_of_filters != number_of_output_filters:
            input = Conv3D(filters=number_of_output_filters,
                            kernel_size=(1, 1, 1),
                            padding='same')(input)

        output = skip_connection(input, output)

        return(output)

    def bottle_neck_block_3d(input, number_of_filters, downsample=False,
                             upsample=False, deconvolution_kernel_size=(2, 2, 2),
                             weight_decay=0.0, dropout_rate=0.0):

        output = input

        number_of_output_filters = number_of_filters

        if downsample:
            output = BatchNormalization()(output)
            output = ThresholdedReLU(theta = 0)(output)

            output = Conv3D(filters=number_of_filters,
                            kernel_size=(1, 1, 1),
                            strides=(2, 2, 2),
                            kernel_initializer=initializers.he_normal(),
                            kernel_regularizer=regularizers.l2(weight_decay))(output)

        output = BatchNormalization()(output)
        output = ThresholdedReLU(theta = 0)(output)

        output = Conv3D(filters=number_of_filters,
                        kernel_size=(1, 1, 1),
                        kernel_initializer=initializers.he_normal(),
                        kernel_regularizer=regularizers.l2(weight_decay))(output)

        output = BatchNormalization()(output)
        output = ThresholdedReLU(theta = 0)(output)

        if upsample:
            output = Conv3DTranspose(filters=number_of_filters,
                                     kernel_size=deconvolution_kernel_size,
                                     padding='same',
                                     kernel_initializer=initializers.he_normal(),
                                     kernel_regularizer=regularizers.l2(weight_decay))(output)
            output = UpSampling3D(size=(2, 2, 2))(output)

        output = Conv3D(filters=(number_of_filters * 4),
                        kernel_size=(1, 1, 1),
                        kernel_initializer=initializers.he_normal(),
                        kernel_regularizer=regularizers.l2(weight_decay))(output)

        number_of_output_filters = number_of_filters * 4

        if dropout_rate > 0.0:
           output=Dropout(rate=dropout_rate)(output)

        # Modify the input so that it has the same size as the output

        if downsample:
            input = Conv3D(filters=number_of_output_filters,
                            kernel_size=(1, 1, 1),
                            strides=(2, 2, 2),
                            padding='same')(input)
        elif upsample:
            input = Conv3DTranspose(filters=number_of_output_filters,
                                    kernel_size=(1, 1, 1),
                                    padding='same')(input)
            input = UpSampling3D(size=(2, 2, 2))(input)
        elif number_of_filters != number_of_output_filters:
            input = Conv3D(filters=number_of_output_filters,
                            kernel_size=(1, 1, 1),
                            padding='valid')(input)

        output = skip_connection(input, output)

        return(output)

    def skip_connection(source, target, merge_mode='sum'):
        layer_list = [source, target]

        output = None
        if merge_mode == 'sum':
            output = Add()(layer_list)
        else:
            channel_axis = 0
            if K.image_data_format() == 'channels_last':
                channel_axis = -1
            output = Concatenate(axis=channel_axis)(layer_list)

        return(output)

    inputs = Input(shape = input_image_size)

    encoding_layers_with_long_skip_connections = []
    encoding_layer_count = 1

    # Preprocessing layer

    model = Conv3D(filters=number_of_filters_at_base_layer,
                   kernel_size=convolution_kernel_size,
                   activation='relu',
                   padding='same',
                   kernel_initializer=initializers.he_normal(),
                   kernel_regularizer=regularizers.l2(weight_decay))(inputs)

    encoding_layers_with_long_skip_connections.append(model)
    encoding_layer_count += 1

    # Encoding initialization path

    model = simple_block_3d(model, number_of_filters_at_base_layer, downsample=True,
                            convolution_kernel_size=convolution_kernel_size,
                            deconvolution_kernel_size=deconvolution_kernel_size,
                            weight_decay=weight_decay, dropout_rate=dropout_rate)

    encoding_layers_with_long_skip_connections.append(model)
    encoding_layer_count += 1

    # Encoding main path

    number_of_bottle_neck_layers = len(bottle_neck_block_depth_schedule)
    for i in range(number_of_bottle_neck_layers):
        number_of_filters = number_of_filters_at_base_layer * 2**i

        for j in range(bottle_neck_block_depth_schedule[i]):

            do_downsample = False
            if j == 0:
                do_downsample = True
            else:
                do_downsample = False

            model = bottle_neck_block_3d(model, number_of_filters=number_of_filters,
                                         downsample=do_downsample,
                                         deconvolution_kernel_size=deconvolution_kernel_size,
                                         weight_decay=weight_decay, dropout_rate=dropout_rate)

            if j == (bottle_neck_block_depth_schedule[i] - 1):
               encoding_layers_with_long_skip_connections.append(model)
               encoding_layer_count += 1

    encoding_layer_count -= 1

    # Transition path

    number_of_filters = number_of_filters_at_base_layer * 2**number_of_bottle_neck_layers

    model = bottle_neck_block_3d(model, number_of_filters=number_of_filters,
                                 downsample=True,
                                 deconvolution_kernel_size=deconvolution_kernel_size,
                                 weight_decay=weight_decay, dropout_rate=dropout_rate)
    model = bottle_neck_block_3d(model, number_of_filters=number_of_filters,
                                 upsample=True,
                                 deconvolution_kernel_size=deconvolution_kernel_size,
                                 weight_decay=weight_decay, dropout_rate=dropout_rate)

    # Decoding main path

    number_of_bottle_neck_layers = len(bottle_neck_block_depth_schedule)
    for i in range(number_of_bottle_neck_layers):
        number_of_filters = (number_of_filters_at_base_layer *
                             2**(number_of_bottle_neck_layers - i - 1))

        for j in range(bottle_neck_block_depth_schedule[number_of_bottle_neck_layers - i - 1]):

            do_upsample = False
            if j == bottle_neck_block_depth_schedule[number_of_bottle_neck_layers - i - 1] - 1:
                do_upsample = True
            else:
                do_upsample = False

            model = bottle_neck_block_3d(model, number_of_filters=number_of_filters,
                                         upsample=do_upsample,
                                         deconvolution_kernel_size=deconvolution_kernel_size,
                                         weight_decay=weight_decay, dropout_rate=dropout_rate)

            if j == 0:
               model = Conv3D(filters=(number_of_filters * 4),
                              kernel_size=(1, 1, 1),
                              padding='same')(model)
               model = skip_connection(encoding_layers_with_long_skip_connections[encoding_layer_count - 1], model)
               encoding_layer_count -= 1

    # Decoding initialization path

    model = simple_block_3d(model, number_of_filters_at_base_layer, upsample=True,
                            convolution_kernel_size=convolution_kernel_size,
                            deconvolution_kernel_size=deconvolution_kernel_size,
                            weight_decay=weight_decay, dropout_rate=dropout_rate)

    # Postprocessing layer

    model = Conv3D(filters=number_of_filters_at_base_layer,
                   kernel_size=convolution_kernel_size,
                   activation='relu',
                   padding='same',
                   kernel_initializer=initializers.he_normal(),
                   kernel_regularizer=regularizers.l2(weight_decay))(model)
    encoding_layer_count -= 1

    model = skip_connection(encoding_layers_with_long_skip_connections[encoding_layer_count - 1], model)

    model = BatchNormalization()(model)
    model = ThresholdedReLU(theta = 0)(model)

    convActivation = ''

    if mode == 'classification':
        convActivation = 'softmax'
    elif mode == 'regression':
        convActivation = 'linear'
    else:
        raise ValueError('mode must be either `classification` or `regression`.')

    outputs = Conv3D(filters=number_of_outputs,
                     kernel_size=(1, 1, 1),
                     activation = convActivation,
                     kernel_regularizer=regularizers.l2(weight_decay))(model)

    resunet_model = Model(inputs=inputs, outputs=outputs)

    return resunet_model




import numpy as np
import ants

def encode_unet(segmentations_array,
                segmentation_labels=None):

    """
    Basic one-hot transformation of segmentations array

    Arguments
    ---------
    segmentations_array : numpy array
        multi-label numpy array

    segmentation_labels : tuple or list
        Note that a background label (typically 0) needs to be included.

    Returns
    -------
    An n-d array of shape batch_size x width x height x <depth> x number_of_segmentation_labels


    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> seg = ants.kmeans_segmentation(image, 3)['segmentation']
    >>> one_hot = encode_unet(seg.numpy().astype('int'))
    """

    if segmentation_labels is None:
        segmentation_labels = np.unique(segmentations_array)

    number_of_labels = len(segmentation_labels)

    dim_segmentations = segmentations_array.shape

    image_dimension = 2
    if len(dim_segmentations) == 4:
        image_dimension = 3

    if number_of_labels < 2:
        raise ValueError("At least two segmentation labels need to be specified.")

    one_hot_array = np.zeros((*dim_segmentations, number_of_labels))
    for i in range(number_of_labels):
        per_label = np.zeros_like(segmentations_array)
        per_label[segmentations_array == segmentation_labels[i]] = 1
        if image_dimension == 2:
            one_hot_array[:,:,:,i] = per_label
        else:
            one_hot_array[:,:,:,:,i] = per_label

    return one_hot_array


def decode_unet(y_predicted,
                domain_image):

    """
    Decoding function for the u-net prediction outcome

    Arguments
    ---------
    y_predicted : an array
        Shape batch_size x width x height x <depth> x number_of_segmentation_labels

    domain_image : ANTs image
        Defines the geometry of the returned probability images

    Returns
    -------
    List of probability images.

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    """

    batch_size = y_predicted.shape[0]
    number_of_labels = y_predicted.shape[-1]

    image_dimension = 2
    if len(y_predicted.shape) == 5:
        image_dimension = 3

    batch_probability_images = list()
    for i in range(batch_size):
        probability_images = list()
        for j in range(number_of_labels):
            if image_dimension == 2:
                image_array = np.squeeze(y_predicted[i,:,:,j])
            else:
                image_array = np.squeeze(y_predicted[i,:,:,:,j])
            probability_images.append(ants.from_numpy(image_array,
               origin=domain_image.origin, spacing=domain_image.spacing,
               direction=domain_image.direction))
        batch_probability_images.append(probability_images)

    return batch_probability_images





import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Add, Activation, BatchNormalization, Concatenate, ReLU, LeakyReLU,
                                     Conv3D, Conv3DTranspose, Input, Lambda, MaxPooling3D,
                                     SpatialDropout3D, UpSampling3D,
                                     Cropping2D, Conv2D, MaxPooling2D, UpSampling2D, ZeroPadding2D)

from tensorflow.keras.activations import softmax

import keras

from ..utilities import InstanceNormalization


def create_nobrainer_unet_model_3d(input_image_size):
    """
    Implementation of the "NoBrainer" U-net architecture

    Creates a keras model of the U-net deep learning architecture for image
    segmentation available at:

            https://github.com/neuronets/nobrainer/


    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    Returns
    -------
    Keras model
        A 3-D keras model defining the U-net network.

    Example
    -------
    >>> model = create_nobrainer_unet_model_3d((None, None, None, 1))
    >>> model.summary()
    """

    number_of_outputs = 1
    number_of_filters_at_base_layer = 16
    convolution_kernel_size = (3, 3, 3)
    deconvolution_kernel_size = (2, 2, 2)

    inputs = Input(shape=input_image_size)

    # Encoding path

    outputs = Conv3D(filters=number_of_filters_at_base_layer,
                     kernel_size=convolution_kernel_size,
                     padding='same')(inputs)
    outputs = ReLU()(outputs)
    outputs = Conv3D(filters=number_of_filters_at_base_layer*2,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)
    skip1 = outputs
    outputs = MaxPooling3D(pool_size=(2, 2, 2))(outputs)

    outputs = Conv3D(filters=number_of_filters_at_base_layer*2,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)
    outputs = Conv3D(filters=number_of_filters_at_base_layer*4,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)
    skip2 = outputs
    outputs = MaxPooling3D(pool_size=(2, 2, 2))(outputs)

    outputs = Conv3D(filters=number_of_filters_at_base_layer*4,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)
    outputs = Conv3D(filters=number_of_filters_at_base_layer*8,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)
    skip3 = outputs
    outputs = MaxPooling3D(pool_size=(2, 2, 2))(outputs)

    outputs = Conv3D(filters=number_of_filters_at_base_layer*8,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)
    outputs = Conv3D(filters=number_of_filters_at_base_layer*16,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)

    # Decoding path

    outputs = Conv3DTranspose(filters=number_of_filters_at_base_layer*16,
                     kernel_size=deconvolution_kernel_size,
                     strides=2,
                     padding='same')(outputs)

    outputs = Concatenate()([skip3, outputs])
    outputs = Conv3D(filters=number_of_filters_at_base_layer*8,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)
    outputs = Conv3D(filters=number_of_filters_at_base_layer*8,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)

    outputs = Conv3DTranspose(filters=number_of_filters_at_base_layer*8,
                     kernel_size=deconvolution_kernel_size,
                     strides=2,
                     padding='same')(outputs)

    outputs = Concatenate()([skip2, outputs])
    outputs = Conv3D(filters=number_of_filters_at_base_layer*4,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)
    outputs = Conv3D(filters=number_of_filters_at_base_layer*4,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)

    outputs = Conv3DTranspose(filters=number_of_filters_at_base_layer*4,
                     kernel_size=deconvolution_kernel_size,
                     strides=2,
                     padding='same')(outputs)

    outputs = Concatenate()([skip1, outputs])
    outputs = Conv3D(filters=number_of_filters_at_base_layer*2,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)
    outputs = Conv3D(filters=number_of_filters_at_base_layer*2,
                     kernel_size=convolution_kernel_size,
                     padding='same')(outputs)
    outputs = ReLU()(outputs)

    convActivation = ''
    if number_of_outputs == 1:
        convActivation = 'sigmoid'
    else:
        convActivation = 'softmax'

    outputs = Conv3D(filters=number_of_outputs,
                     kernel_size=1,
                     activation = convActivation)(outputs)

    unet_model = Model(inputs=inputs, outputs=outputs)

    return unet_model

def create_hippmapp3r_unet_model_3d(input_image_size,
                                    do_first_network=True,
                                    data_format="channels_last"):
    """
    Implementation of the "HippMapp3r" U-net architecture

    Creates a keras model implementation of the u-net architecture
    described here:

        https://onlinelibrary.wiley.com/doi/pdf/10.1002/hbm.24811

    with the implementation available here:

        https://github.com/mgoubran/HippMapp3r

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    do_first_network : boolean
        Boolean dictating if the model built should be the first (initial) network
        or second (refinement) network.

    data_format : string
        One of "channels_first" or "channels_last".  We do this for this specific
        architecture as the original weights were saved in "channels_first" format.

    Returns
    -------
    Keras model
        A 3-D keras model defining the U-net network.

    Example
    -------
    >>> shape_initial_stage = (160, 160, 128)
    >>> model_initial_stage = antspynet.create_hippmapp3r_unet_model_3d((*shape_initial_stage, 1), True)
    >>> model_initial_stage.load_weights(antspynet.get_pretrained_network("hippMapp3rInitial"))
    >>> shape_refine_stage = (112, 112, 64)
    >>> model_refine_stage = antspynet.create_hippmapp3r_unet_model_3d((*shape_refine_stage, 1), False)
    >>> model_refine_stage.load_weights(antspynet.get_pretrained_network("hippMapp3rRefine"))
    """

    channels_axis = None
    if data_format == "channels_last":
        channels_axis = 4
    elif data_format == "channels_first":
        channels_axis = 1
    else:
        raise ValueError("Unexpected string for data_format.")

    def convB_3d_layer(input, number_of_filters, kernel_size=3, strides=1):
        block = Conv3D(filters=number_of_filters,
                       kernel_size=kernel_size,
                       strides=strides,
                       padding='same',
                       data_format=data_format)(input)
        block = InstanceNormalization(axis=channels_axis)(block)
        block = LeakyReLU()(block)
        return(block)

    def residual_block_3d(input, number_of_filters):
        block = convB_3d_layer(input, number_of_filters)
        block = SpatialDropout3D(rate=0.3,
                                 data_format=data_format)(block)
        block = convB_3d_layer(block, number_of_filters)
        return(block)

    def upsample_block_3d(input, number_of_filters):
        block = UpSampling3D(data_format=data_format)(input)
        block = convB_3d_layer(block, number_of_filters)
        return(block)

    def feature_block_3d(input, number_of_filters):
        block = convB_3d_layer(input, number_of_filters)
        block = convB_3d_layer(block, number_of_filters, kernel_size=1)
        return(block)

    number_of_filters_at_base_layer = 16

    number_of_layers = 6
    if do_first_network == False:
        number_of_layers = 5

    inputs = Input(shape=input_image_size)

    # Encoding path

    add = None

    encoding_convolution_layers = []
    for i in range(number_of_layers):
        number_of_filters = number_of_filters_at_base_layer * 2 ** i
        conv = None
        if i == 0:
            conv = convB_3d_layer(inputs, number_of_filters)
        else:
            conv = convB_3d_layer(add, number_of_filters, strides=2)
        residual_block = residual_block_3d(conv, number_of_filters)
        add = Add()([conv, residual_block])
        encoding_convolution_layers.append(add)

    # Decoding path

    outputs = encoding_convolution_layers[number_of_layers-1]

    # 256
    number_of_filters = (number_of_filters_at_base_layer *
      2 ** (number_of_layers - 2))
    outputs = upsample_block_3d(outputs, number_of_filters)

    if do_first_network == True:
        # 256, 128
        outputs = Concatenate(axis=channels_axis)([encoding_convolution_layers[4], outputs])
        outputs = feature_block_3d(outputs, number_of_filters)
        number_of_filters = int(number_of_filters / 2)
        outputs = upsample_block_3d(outputs, number_of_filters)

    # 128, 64
    outputs = Concatenate(axis=channels_axis)([encoding_convolution_layers[3], outputs])
    outputs = feature_block_3d(outputs, number_of_filters)
    number_of_filters = int(number_of_filters / 2)
    outputs = upsample_block_3d(outputs, number_of_filters)

    # 64, 32
    outputs = Concatenate(axis=channels_axis)([encoding_convolution_layers[2], outputs])
    feature64 = feature_block_3d(outputs, number_of_filters)
    number_of_filters = int(number_of_filters / 2)
    outputs = upsample_block_3d(feature64, number_of_filters)
    back64 = None
    if do_first_network == True:
        back64 = convB_3d_layer(feature64, 1, 1)
    else:
        back64 = Conv3D(filters=1,
                        kernel_size=1,
                        data_format=data_format)(feature64)
    back64 = UpSampling3D(data_format=data_format)(back64)

    # 32, 16
    outputs = Concatenate(axis=channels_axis)([encoding_convolution_layers[1], outputs])
    feature32 = feature_block_3d(outputs, number_of_filters)
    number_of_filters = int(number_of_filters / 2)
    outputs = upsample_block_3d(feature32, number_of_filters)
    back32 = None
    if do_first_network == True:
        back32 = convB_3d_layer(feature32, 1, 1)
    else:
        back32 = Conv3D(filters=1,
                        kernel_size=1,
                        data_format=data_format)(feature32)
    back32 = Add()([back64, back32])
    back32 = UpSampling3D(data_format=data_format)(back32)

    # Final
    outputs = Concatenate(axis=channels_axis)([encoding_convolution_layers[0], outputs])
    outputs = convB_3d_layer(outputs, number_of_filters, 3)
    outputs = convB_3d_layer(outputs, number_of_filters, 1)
    if do_first_network == True:
        outputs = convB_3d_layer(outputs, 1, 1)
    else:
        outputs = Conv3D(filters=1,
                         kernel_size=1,
                         data_format=data_format)(outputs)
    outputs = Add()([back32, outputs])
    outputs = Activation('sigmoid')(outputs)

    unet_model = Model(inputs=inputs, outputs=outputs)

    return(unet_model)

def create_sysu_media_unet_model_2d(input_image_size,
                                    anatomy="wmh"):
    """
    Implementation of the sysu_media U-net architecture

    Creates a keras model implementation of the u-net architecture
    in the 2017 MICCAI WMH challenge by the sysu_medial team described
    here:

        https://pubmed.ncbi.nlm.nih.gov/30125711/

    with the original implementation available here:

        https://github.com/hongweilibran/wmh_ibbmTum

    Arguments
    ---------
    input_image_size : tuple of length 4
        This will be (200, 200, 2) for t1/flair input and (200, 200, 1)} for
        flair-only input.

    anatomy : string
        "wmh" or "claustrum"

    Returns
    -------
    Keras model
        A 2-D keras model defining the U-net network.

    Example
    -------
    >>> image_size = (200, 200)
    >>> model = antspynet.create_sysu_media_unet_model_2d((*image_size, 1))
    """

    def get_crop_shape(target_layer, reference_layer):

        delta = K.int_shape(target_layer)[1] - K.int_shape(reference_layer)[1]
        if delta % 2 != 0:
            cropShape0 = (int(delta/2), int(delta/2) + 1)
        else:
            cropShape0 = (int(delta/2), int(delta/2))

        delta = K.int_shape(target_layer)[2] - K.int_shape(reference_layer)[2]
        if delta % 2 != 0:
            cropShape1 = (int(delta/2), int(delta/2) + 1)
        else:
            cropShape1 = (int(delta/2), int(delta/2))

        return((cropShape0, cropShape1))

    inputs = Input(shape=input_image_size)

    if anatomy == "wmh":
        number_of_filters = (64, 96, 128, 256, 512)
    elif anatomy == "claustrum":
        number_of_filters = (32, 64, 96, 128, 256)

    # encoding layers

    encoding_layers = list()

    outputs = inputs
    for i in range(len(number_of_filters)):

        kernel1 = 3
        kernel2 = 3
        if i == 0 and anatomy == "wmh":
            kernel1 = 5
            kernel2 = 5
        elif i == 3:
            kernel1 = 3
            kernel2 = 4

        outputs = Conv2D(filters=number_of_filters[i],
                         kernel_size=kernel1,
                         padding='same')(outputs)
        outputs = Activation('relu')(outputs)
        outputs = Conv2D(filters=number_of_filters[i],
                         kernel_size=kernel2,
                         padding='same')(outputs)
        outputs = Activation('relu')(outputs)
        encoding_layers.append(outputs)
        if i < 4:
            outputs = MaxPooling2D(pool_size=(2, 2))(outputs)

    # decoding layers

    for i in range(len(encoding_layers)-2, -1, -1):
        upsample_layer = UpSampling2D(size=(2, 2))(outputs)
        crop_shape = get_crop_shape(encoding_layers[i], upsample_layer)
        cropped_layer = Cropping2D(cropping=crop_shape)(encoding_layers[i])
        outputs = Concatenate(axis=-1)([upsample_layer, cropped_layer])
        outputs = Conv2D(filters=number_of_filters[i],
                         kernel_size=3,
                         padding='same')(outputs)
        outputs = Activation('relu')(outputs)
        outputs = Conv2D(filters=number_of_filters[i],
                         kernel_size=3,
                         padding='same')(outputs)
        outputs = Activation('relu')(outputs)

    # final

    crop_shape = get_crop_shape(inputs, outputs)
    outputs = ZeroPadding2D(padding=crop_shape)(outputs)
    outputs = Conv2D(filters=1,
                     kernel_size=1,
                     activation='sigmoid',
                     padding='same')(outputs)

    unet_model = Model(inputs=inputs, outputs=outputs)

    return(unet_model)


def create_hypothalamus_unet_model_3d(input_image_size):

    """
    Implementation of the U-net architecture for hypothalamus segmentation
    described in

    https://pubmed.ncbi.nlm.nih.gov/32853816/

    and ported from the original implementation:

        https://github.com/BBillot/hypothalamus_seg

    The network has is characterized by the following parameters:
        * 3 resolution levels:  24 ---> 48 ---> 96 filters
        * convolution: kernel size:  (3, 3, 3), activation: 'elu',
        * pool size: (2, 2, 2)

    Returns
    -------
    Keras model
        A 3-D keras model defining the U-net network.

    Example
    -------
    >>> model = create_hypothalamus_unet_model_3d((160, 160, 160, 1))
    """

    convolution_kernel_size = (3, 3, 3)
    pool_size = (2, 2, 2)
    number_of_outputs = 11

    number_of_layers = 3
    number_of_filters_at_base_layer = 24
    number_of_filters = list()
    for i in range(number_of_layers):
        number_of_filters.append(number_of_filters_at_base_layer * 2**i)

    inputs = Input(shape=(*input_image_size, 1))

    # Encoding path

    encoding_convolution_layers = []
    pool = None
    for i in range(number_of_layers):

        if i == 0:
            conv = Conv3D(filters=number_of_filters[i],
                          kernel_size=convolution_kernel_size,
                          padding='same',
                          activation='elu')(inputs)
        else:
            conv = Conv3D(filters=number_of_filters[i],
                          kernel_size=convolution_kernel_size,
                          padding='same',
                          activation='elu')(pool)

        conv = Conv3D(filters=number_of_filters[i],
                      kernel_size=convolution_kernel_size,
                      padding='same',
                      activation='elu')(conv)

        encoding_convolution_layers.append(conv)

        conv = BatchNormalization(axis=-1)(conv)

        if i < number_of_layers - 1:
            pool = MaxPooling3D(pool_size=pool_size)(conv)
        else:
            outputs = conv

    # Decoding path

    for i in range(1, number_of_layers):

        deconv = UpSampling3D(size=pool_size)(outputs)
        outputs = Concatenate(axis=4)([encoding_convolution_layers[number_of_layers-i-1], deconv])
        outputs = Conv3D(filters=number_of_filters[number_of_layers-i-1],
                         kernel_size=convolution_kernel_size,
                         padding='same',
                         activation='elu')(outputs)
        outputs = Conv3D(filters=number_of_filters[number_of_layers-i-1],
                         kernel_size=convolution_kernel_size,
                         padding='same',
                         activation='elu')(outputs)

        outputs = BatchNormalization(axis=-1)(outputs)

    outputs = Conv3D(filters=number_of_outputs,
                     kernel_size=(1, 1, 1),
                     activation='softmax')(outputs)

    unet_model = Model(inputs=inputs, outputs=outputs)

    return(unet_model)


import tensorflow as tf

def get_pretrained_network(file_id=None,
                           target_file_name=None,
                           antsxnet_cache_directory=None):

    """
    Download pretrained network/weights.

    Arguments
    ---------

    file_id string
        One of the permitted file ids or pass "show" to list all
        valid possibilities. Note that most require internet access
        to download.

    target_file_name string
       Optional target filename.

    antsxnet_cache_directory string
       Optional target output.  If not specified these data will be downloaded
       to the subdirectory ~/.keras/ANTsXNet/.

    Returns
    -------
    A filename string

    Example
    -------
    >>> model_file = get_pretrained_network('dbpn4x')
    """

    def switch_networks(argument):
        switcher = {
            "brainAgeGender": "https://ndownloader.figshare.com/files/22179948",
            "brainAgeDeepBrainNet": "https://ndownloader.figshare.com/files/23573402",
            "brainExtraction": "https://ndownloader.figshare.com/files/22944632",
            "brainExtractionT1": "https://ndownloader.figshare.com/files/27334370",
            "brainExtractionT1v1": "https://ndownloader.figshare.com/files/28057626",
            "brainExtractionT2": "https://ndownloader.figshare.com/files/23066153",
            "brainExtractionFLAIR": "https://ndownloader.figshare.com/files/23562194",
            "brainExtractionBOLD": "https://ndownloader.figshare.com/files/22761977",
            "brainExtractionFA": "https://ndownloader.figshare.com/files/22761926",
            "brainExtractionNoBrainer": "https://ndownloader.figshare.com/files/22598039",
            "brainExtractionInfantT1T2": "https://ndownloader.figshare.com/files/22968833",
            "brainExtractionInfantT1": "https://ndownloader.figshare.com/files/22968836",
            "brainExtractionInfantT2": "https://ndownloader.figshare.com/files/22968830",
            "brainSegmentation": "https://ndownloader.figshare.com/files/13900010",
            "brainSegmentationPatchBased": "https://ndownloader.figshare.com/files/14249717",
            "claustrum_axial_0": "https://ndownloader.figshare.com/files/27844068",
            "claustrum_axial_1": "https://ndownloader.figshare.com/files/27844059",
            "claustrum_axial_2": "https://ndownloader.figshare.com/files/27844062",
            "claustrum_corona_0": "https://ndownloader.figshare.com/files/27844074",
            "claustrum_corona_1": "https://ndownloader.figshare.com/files/27844071",
            "claustrum_corona_2": "https://ndownloader.figshare.com/files/27844065",
            "ctHumanLung": "https://ndownloader.figshare.com/files/20005217",
            "dbpn4x": "https://ndownloader.figshare.com/files/13347617",
            "deepFlash": "https://ndownloader.figshare.com/files/22933757",
            "deepFlashLeft8": "https://ndownloader.figshare.com/files/25441007",
            "deepFlashRight8": "https://ndownloader.figshare.com/files/25441004",
            "deepFlashLeft16": "https://ndownloader.figshare.com/files/25465844",
            "deepFlashRight16": "https://ndownloader.figshare.com/files/25465847",
            "deepFlashLeft16new": "https://ndownloader.figshare.com/files/25991681",
            "deepFlashRight16new": "https://ndownloader.figshare.com/files/25991678",
            "denoising": "https://ndownloader.figshare.com/files/14235296",
            "dktInner": "https://ndownloader.figshare.com/files/23266943",
            "dktOuter": "https://ndownloader.figshare.com/files/23765132",
            "dktOuterWithSpatialPriors": "https://ndownloader.figshare.com/files/24230768",
            "elBicho": "https://ndownloader.figshare.com/files/26736779",
            "ewDavidWmhSegmentationWeights": "https://ndownloader.figshare.com/files/26827691",
            # "ewDavidWmhSegmentationSlicewiseWeights": "https://ndownloader.figshare.com/files/26787152",
            "ewDavidWmhSegmentationSlicewiseWeights": "https://ndownloader.figshare.com/files/26896703",
            "ewDavidWmhSegmentationSlicewiseT1OnlyWeights": "https://ndownloader.figshare.com/files/26920130",
            "ewDavidSysu": "",
            "ewDavidSysuT1Only": "",
            "ewDavidSysuFlairOnly": "",
            "ewDavidSysuPlus": "",
            "ewDavidSysuPlusT1Only": "",
            "ewDavidSysuPlusFlairOnly": "",
            "ewDavidSysuPlusSeg": "",
            "ewDavidSysuPlusSegT1Only": "",
            "ewDavidSysuWithSite": "",
            "ewDavidSysuWithSiteT1Only": "",
            "ewDavidSysuWithSiteFlairOnly": "",
            "functionalLungMri": "https://ndownloader.figshare.com/files/13824167",
            "hippMapp3rInitial": "https://ndownloader.figshare.com/files/18068408",
            "hippMapp3rRefine": "https://ndownloader.figshare.com/files/18068411",
            "hypothalamus": "https://ndownloader.figshare.com/files/28344378",
            "koniqMBCS": "https://ndownloader.figshare.com/files/24967376",
            "koniqMS": "https://ndownloader.figshare.com/files/25461887",
            "koniqMS2": "https://ndownloader.figshare.com/files/25474850",
            "koniqMS3": "https://ndownloader.figshare.com/files/25474847",
            "lungCtWithPriorsSegmentationWeights": "https://ndownloader.figshare.com/files/28357818",
            "mriSuperResolution": "https://ndownloader.figshare.com/files/24128618",
            "protonLungMri": "https://ndownloader.figshare.com/files/13606799",
            "sixTissueOctantBrainSegmentation": "https://ndownloader.figshare.com/files/23776025",
            "sixTissueOctantBrainSegmentationWithPriors1": "https://ndownloader.figshare.com/files/28159869",
            "sysuMediaWmhFlairOnlyModel0": "https://ndownloader.figshare.com/files/22898441",
            "sysuMediaWmhFlairOnlyModel1": "https://ndownloader.figshare.com/files/22898570",
            "sysuMediaWmhFlairOnlyModel2": "https://ndownloader.figshare.com/files/22898438",
            "sysuMediaWmhFlairT1Model0": "https://ndownloader.figshare.com/files/22898450",
            "sysuMediaWmhFlairT1Model1": "https://ndownloader.figshare.com/files/22898453",
            "sysuMediaWmhFlairT1Model2": "https://ndownloader.figshare.com/files/22898459",
            "tidsQualityAssessment": "https://ndownloader.figshare.com/files/24292895",
            "wholeTumorSegmentationT2Flair": "https://ndownloader.figshare.com/files/14087045"
        }
        return(switcher.get(argument, "Invalid argument."))

    if file_id == None:
        raise ValueError("Missing file id.")

    valid_list = ("dbpn4x",
                  "brainAgeGender",
                  "brainAgeDeepBrainNet",
                  "brainExtraction",
                  "brainExtractionT1",
                  "brainExtractionT1v1",
                  "brainExtractionT2",
                  "brainExtractionFLAIR",
                  "brainExtractionBOLD",
                  "brainExtractionFA",
                  "brainExtractionNoBrainer",
                  "brainExtractionInfantT1T2",
                  "brainExtractionInfantT1",
                  "brainExtractionInfantT2",
                  "brainSegmentation",
                  "brainSegmentationPatchBased",
                  "claustrum_axial_0",
                  "claustrum_axial_1",
                  "claustrum_axial_2",
                  "claustrum_coronal_0",
                  "claustrum_coronal_1",
                  "claustrum_coronal_2",
                  "ctHumanLung",
                  "deepFlash",
                  "deepFlashLeft8",
                  "deepFlashRight8",
                  "deepFlashLeft16",
                  "deepFlashRight16",
                  "deepFlashLeft16new",
                  "deepFlashRight16new",
                  "denoising",
                  "dktInner",
                  "dktOuter",
                  "dktOuterWithSpatialPriors",
                  "elBicho",
                  "ewDavidWmhSegmentationWeights",
                  "ewDavidWmhSegmentationSlicewiseWeights",
                  "ewDavidWmhSegmentationSlicewiseT1OnlyWeights",
                  "ewDavidSysu",
                  "ewDavidSysuT1Only",
                  "ewDavidSysuFlairOnly",
                  "ewDavidSysuPlus",
                  "ewDavidSysuPlusT1Only",
                  "ewDavidSysuPlusFlairOnly",
                  "ewDavidSysuPlusSeg",
                  "ewDavidSysuPlusSegT1Only",
                  "ewDavidSysuWithSite",
                  "ewDavidSysuWithSiteT1Only",
                  "ewDavidSysuWithSiteFlairOnly",
                  "functionalLungMri",
                  "hippMapp3rInitial",
                  "hippMapp3rRefine",
                  "hypothalamus",
                  "koniqMBCS",
                  "koniqMS",
                  "koniqMS2",
                  "koniqMS3",
                  "lungCtWithPriorsSegmentationWeights",
                  "mriSuperResolution",
                  "protonLungMri",
                  "sixTissueOctantBrainSegmentation",
                  "sixTissueOctantBrainSegmentationWithPriors1",
                  "sixTissueOctantBrainSegmentationWithPriors2",
                  "sysuMediaWmhFlairOnlyModel0",
                  "sysuMediaWmhFlairOnlyModel1",
                  "sysuMediaWmhFlairOnlyModel2",
                  "sysuMediaWmhFlairT1Model0",
                  "sysuMediaWmhFlairT1Model1",
                  "sysuMediaWmhFlairT1Model2",
                  "tidsQualityAssessment",
                  "wholeTumorSegmentationT2Flair",
                  "show")

    if not file_id in valid_list:
        raise ValueError("No data with the id you passed - try \"show\" to get list of valid ids.")

    if file_id == "show":
       return(valid_list)

    url = switch_networks(file_id)

    if target_file_name == None:
        target_file_name = file_id + ".h5"

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    target_file_name_path = tf.keras.utils.get_file(target_file_name, url,
        cache_subdir = antsxnet_cache_directory)

    return(target_file_name_path)



import ants

import numpy as np
import math

def crop_image_center(image,
                      crop_size):
    """
    Crop the center of an image.

    Arguments
    ---------
    image : ANTsImage
        Input image

    crop_size: n-D tuple (depending on dimensionality).
        Width, height, depth (if 3-D), and time (if 4-D) of crop region.

    Returns
    -------
    A list (or array) of patches.

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> cropped_image = crop_image_center(image, crop_size=(64, 64))
    """

    image_size = np.array(image.shape)

    if len(image_size) != len(crop_size):
        raise ValueError("crop_size does not match image size.")

    if (np.asarray(crop_size) > np.asarray(image_size)).any():
        raise ValueError("A crop_size dimension is larger than image_size.")

    start_index = (np.floor(0.5 * (np.asarray(image_size) - np.asarray(crop_size)))).astype(int)
    end_index = start_index + np.asarray(crop_size).astype(int)

    cropped_image = ants.crop_indices(ants.image_clone(image) * 1, start_index, end_index)

    return(cropped_image)

def pad_image_by_factor(image,
                        factor):
    """
    Pad an image based on a factor.

    Pad image of size (x, y, z) to (x', y', z') where (x', y', z')
    is a divisible by a user-specified factor.

    Arguments
    ---------
    image : ANTsImage
        Input image

    factor: scalar or n-D tuple
        Padding factor

    Returns
    -------
    A padded image

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> padded_image = pad_image_by_factor(image, factor=4)
    """

    factor_vector = factor
    if isinstance(factor, int):
        factor_vector = np.repeat(factor, image.dimension)

    if len(factor_vector) != image.dimension:
        raise ValueError("factor must be scalar or the length of the image dimension.")

    image_size = np.array(image.shape)
    delta_size = image_size % factor_vector

    padded_size = image_size
    for i in range(len(padded_size)):
        if delta_size[i] > 0:
            padded_size[i] = image_size[i] - delta_size[i] + factor_vector[i]

    padded_image = pad_or_crop_image_to_size(image, padded_size)

    return(padded_image)

def pad_or_crop_image_to_size(image,
                              size):
    """
    Pad or crop an image to a specified size

    Arguments
    ---------
    image : ANTsImage
        Input image

    size : tuple
        size of output image

    Returns
    -------
    A cropped or padded image

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> padded_image = pad_or_crop_image_to_size(image, (333, 333))
    """

    image_size = np.array(image.shape)

    delta = image_size - np.array(size)

    if np.any(delta < 0):
        pad_size = abs(delta.min())
        pad_shape = image_size + pad_size
        image = ants.pad_image(image, shape=pad_shape)

    cropped_image = crop_image_center(image, size)

    return(cropped_image)





from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Dense

def create_autoencoder_model(number_of_units_per_layer,
                             activation='relu',
                             initializer='glorot_uniform'
                            ):
    """
    2-D implementation of the Vgg deep learning architecture.

    Builds an autoencoder based on the specified array definining the
    number of units in the encoding branch.  Ported to Keras R from the
    Keras python implementation here:

    https://github.com/XifengGuo/DEC-keras

    Arguments
    ---------
    number_of_units_per_layer : tuple
        A tuple defining the number of units in the encoding branch.

    activation : string
        Activation type for the dense layers

    initializer : string
        Initializer type for the dense layers

    Returns
    -------
    Keras model
        An encoder and autoencoder Keras model.

    Example
    -------
    >>> model = create_autoencoder_model((784, 500, 500, 2000, 10))
    >>> model.summary()
    """

    number_of_encoding_layers = len(number_of_units_per_layer) - 1

    inputs = Input(shape=(number_of_units_per_layer[0],))

    encoder = inputs

    for i in range(number_of_encoding_layers - 1):
        encoder = Dense(units=number_of_units_per_layer[i + 1],
                        activation=activation,
                        kernel_initializer=initializer)(encoder)

    encoder = Dense(units=number_of_units_per_layer[-1])(encoder)

    autoencoder = encoder

    for i in range(number_of_encoding_layers-1, 0, -1):
        autoencoder = Dense(units=number_of_units_per_layer[i],
                            activation=activation,
                            kernel_initializer=initializer)(autoencoder)

    autoencoder = Dense(units=number_of_units_per_layer[0],
                        kernel_initializer=initializer)(autoencoder)

    encoder_model = Model(inputs=inputs, outputs=encoder)
    autoencoder_model = Model(inputs=inputs, outputs=autoencoder)

    return(autoencoder_model, encoder_model)



import numpy as np
import tensorflow as tf
import ants

def hippmapp3r_segmentation(t1,
                            do_preprocessing=True,
                            antsxnet_cache_directory=None,
                            verbose=False):

    """
    Perform HippMapp3r (hippocampal) segmentation described in

     https://www.ncbi.nlm.nih.gov/pubmed/31609046

    with models and architecture ported from

    https://github.com/mgoubran/HippMapp3r

    Additional documentation and attribution resources found at

    https://hippmapp3r.readthedocs.io/en/latest/

    Preprocessing consists of:
       * n4 bias correction and
       * brain extraction
    The input T1 should undergo the same steps.  If the input T1 is the raw
    T1, these steps can be performed by the internal preprocessing, i.e. set
    do_preprocessing = True

    Arguments
    ---------
    t1 : ANTsImage
        input image

    do_preprocessing : boolean
        See description above.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    ANTs labeled hippocampal image.

    Example
    -------
    >>> mask = hippmapp3r_segmentation(t1)
    """

    from ..architectures import create_hippmapp3r_unet_model_3d
    from ..utilities import preprocess_brain_image
    from ..utilities import get_pretrained_network
    from ..utilities import get_antsxnet_data

    if t1.dimension != 3:
        raise ValueError( "Image dimension must be 3." )

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    if verbose == True:
        print("*************  Preprocessing  ***************")
        print("")

    t1_preprocessed = t1
    if do_preprocessing == True:
        t1_preprocessing = preprocess_brain_image(t1,
            truncate_intensity=None,
            brain_extraction_modality="t1",
            template=None,
            do_bias_correction=True,
            do_denoising=False,
            antsxnet_cache_directory=antsxnet_cache_directory,
            verbose=verbose)
        t1_preprocessed = t1_preprocessing["preprocessed_image"] * t1_preprocessing['brain_mask']

    if verbose == True:
        print("*************  Initial stage segmentation  ***************")
        print("")

    # Normalize to mprage_hippmapp3r space
    if verbose == True:
        print("    HippMapp3r: template normalization.")

    template_file_name_path = get_antsxnet_data("mprage_hippmapp3r", antsxnet_cache_directory=antsxnet_cache_directory)
    template_image = ants.image_read(template_file_name_path)

    registration = ants.registration(fixed=template_image, moving=t1_preprocessed,
        type_of_transform="antsRegistrationSyNQuickRepro[t]", verbose=verbose)
    image = registration['warpedmovout']
    transforms = dict(fwdtransforms=registration['fwdtransforms'],
                        invtransforms=registration['invtransforms'])

    # Threshold at 10th percentile of non-zero voxels in "robust range (fslmaths)"
    if verbose == True:
        print("    HippMapp3r: threshold.")

    image_array = image.numpy()
    image_robust_range = np.quantile(image_array[np.where(image_array != 0)], (0.02, 0.98))
    threshold_value = 0.10 * (image_robust_range[1] - image_robust_range[0]) + image_robust_range[0]
    thresholded_mask = ants.threshold_image(image, -10000, threshold_value, 0, 1)
    thresholded_image = image * thresholded_mask

    # Standardize image
    if verbose == True:
        print("    HippMapp3r: standardize.")

    mean_image = np.mean(thresholded_image[thresholded_mask==1])
    sd_image = np.std(thresholded_image[thresholded_mask==1])
    image_normalized = (image - mean_image) / sd_image
    image_normalized = image_normalized * thresholded_mask

    # Trim and resample image
    if verbose == True:
        print("    HippMapp3r: trim and resample to (160, 160, 128).")

    image_cropped = ants.crop_image(image_normalized, thresholded_mask, 1)
    shape_initial_stage = (160, 160, 128)
    image_resampled = ants.resample_image(image_cropped, shape_initial_stage, use_voxels=True, interp_type=1)

    if verbose == True:
        print("    HippMapp3r: generate first network and download weights.")

    model_initial_stage = create_hippmapp3r_unet_model_3d((*shape_initial_stage, 1), do_first_network=True)

    initial_stage_weights_file_name = get_pretrained_network("hippMapp3rInitial", antsxnet_cache_directory=antsxnet_cache_directory)
    model_initial_stage.load_weights(initial_stage_weights_file_name)

    if verbose == True:
        print("    HippMapp3r: prediction.")

    data_initial_stage = np.expand_dims(image_resampled.numpy(), axis=0)
    data_initial_stage = np.expand_dims(data_initial_stage, axis=-1)
    mask_array = model_initial_stage.predict(data_initial_stage, verbose=verbose)
    mask_image_resampled = ants.copy_image_info(image_resampled, ants.from_numpy(np.squeeze(mask_array)))
    mask_image = ants.resample_image(mask_image_resampled, image.shape, use_voxels=True, interp_type=0)
    mask_image[mask_image >= 0.5] = 1
    mask_image[mask_image < 0.5] = 0

    #########################################
    #
    # Perform refined (stage 2) segmentation
    #

    if verbose == True:
        print("")
        print("")
        print("*************  Refine stage segmentation  ***************")
        print("")

    mask_array = np.squeeze(mask_array)
    centroid_indices = np.where(mask_array == 1)
    centroid = np.zeros((3,))
    centroid[0] = centroid_indices[0].mean()
    centroid[1] = centroid_indices[1].mean()
    centroid[2] = centroid_indices[2].mean()

    shape_refine_stage = (112, 112, 64)
    lower = (np.floor(centroid - 0.5 * np.array(shape_refine_stage)) - 1).astype(int)
    upper = (lower + np.array(shape_refine_stage)).astype(int)

    image_trimmed = ants.crop_indices(image_resampled, lower.astype(int), upper.astype(int))

    if verbose == True:
        print("    HippMapp3r: generate second network and download weights.")

    model_refine_stage = create_hippmapp3r_unet_model_3d((*shape_refine_stage, 1), do_first_network=False)

    refine_stage_weights_file_name = get_pretrained_network("hippMapp3rRefine", antsxnet_cache_directory=antsxnet_cache_directory)
    model_refine_stage.load_weights(refine_stage_weights_file_name)

    data_refine_stage = np.expand_dims(image_trimmed.numpy(), axis=0)
    data_refine_stage = np.expand_dims(data_refine_stage, axis=-1)

    if verbose == True:
        print("    HippMapp3r: o Monte Carlo iterations (SpatialDropout).")

    number_of_mci_iterations = 30
    prediction_refine_stage = np.zeros(shape_refine_stage)
    for i in range(number_of_mci_iterations):
        tf.random.set_seed(i)
        if verbose == True:
            print("        Monte Carlo iteration", i + 1, "out of", number_of_mci_iterations)
        prediction_refine_stage = \
            (np.squeeze(model_refine_stage.predict(data_refine_stage, verbose=verbose)) + \
             i * prediction_refine_stage ) / (i + 1)

    prediction_refine_stage_array = np.zeros(image_resampled.shape)
    prediction_refine_stage_array[lower[0]:upper[0],
                                  lower[1]:upper[1],
                                  lower[2]:upper[2]] = prediction_refine_stage
    probability_mask_refine_stage_resampled = ants.copy_image_info(image_resampled, ants.from_numpy(prediction_refine_stage_array))

    segmentation_image_resampled = ants.label_clusters(
        ants.threshold_image(probability_mask_refine_stage_resampled, 0.0, 0.5, 0, 1), min_cluster_size=10)
    segmentation_image_resampled[segmentation_image_resampled > 2] = 0
    geom = ants.label_geometry_measures(segmentation_image_resampled)
    if len(geom['VolumeInMillimeters']) < 2:
        raise ValueError("Error: left and right hippocampus not found.")

    if geom['Centroid_x'][0] < geom['Centroid_x'][1]:
        segmentation_image_resampled[segmentation_image_resampled == 1] = 3
        segmentation_image_resampled[segmentation_image_resampled == 2] = 1
        segmentation_image_resampled[segmentation_image_resampled == 3] = 2

    segmentation_image = ants.apply_transforms(fixed=t1,
      moving=segmentation_image_resampled, transformlist=transforms['invtransforms'],
      whichtoinvert=[True], interpolator="genericLabel", verbose=verbose)

    return(segmentation_image)



from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Concatenate, Dense, Dropout, Add, Activation,
                                     multiply, ReLU, LeakyReLU,
                                     GlobalAveragePooling2D,
                                     Conv2D, Conv2DTranspose, MaxPooling2D,
                                     UpSampling2D,
                                     GlobalAveragePooling3D,
                                     Conv3D, Conv3DTranspose, MaxPooling3D,
                                     UpSampling3D)
from tensorflow.keras import regularizers

from ..utilities import InstanceNormalization

def create_unet_model_2d(input_image_size,
                         number_of_outputs=2,
                         scalar_output_size = 0,
                         scalar_output_activation = "relu",
                         number_of_layers=4,
                         number_of_filters_at_base_layer=32,
                         number_of_filters=None,
                         convolution_kernel_size=(3, 3),
                         deconvolution_kernel_size=(2, 2),
                         pool_size=(2, 2),
                         strides=(2, 2),
                         dropout_rate=0.0,
                         weight_decay=0.0,
                         mode='classification',
                         additional_options=None
                        ):
    """
    2-D implementation of the U-net deep learning architecture.

    Creates a keras model of the U-net deep learning architecture for image
    segmentation and regression.  More information is provided at the authors'
    website:

            https://lmb.informatik.uni-freiburg.de/people/ronneber/u-net/

    with the paper available here:

            https://arxiv.org/abs/1505.04597

    This particular implementation was influenced by the following python
    implementation:

            https://github.com/joelthelion/ultrasound-nerve-segmentation


    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    number_of_outputs : integer
        Meaning depends on the mode.  For `classification` this is the number of
        segmentation labels.  For `regression` this is the number of outputs.

    scalar_output_size : integer
        If greater than 0, a global average pooling from each
        encoding layer is concatenated to a dense layer as a secondary output.

    scalar_output_activation : string
        Activation for nonzero output scalar.

    number_of_layers : integer
        number of encoding/decoding layers.

    number_of_filters_at_base_layer : integer
        number of filters at the beginning and end of the `U`.  Doubles at each
        descending/ascending layer.

    number_of_filters : tuple
        tuple explicitly setting the number of filters at each layer.  One can
        either set this or number_of_layers and  number_of_filters_at_base_layer.
        Default = None.

    convolution_kernel_size : tuple of length 2
        Defines the kernel size during the encoding.

    deconvolution_kernel_size : tuple of length 2
        Defines the kernel size during the decoding.

    pool_size : tuple of length 2
        Defines the region for each pooling layer.

    strides : tuple of length 2
        Strides for the convolutional layers.

    dropout_rate : scalar
        Float between 0 and 1 to use between dense layers.

    weight_decay :  scalar
        Weighting parameter for L2 regularization of the kernel weights of the
        convolution layers.  Default = 0.0.

    mode :  string
        `classification`, `regression`, or `sigmoid`.  Default = `classification`.

    additional_options : string or tuple of strings
        specific configuration add-ons/tweaks:
            * "attentionGating" -- attention-unet variant in https://pubmed.ncbi.nlm.nih.gov/33288961/
            * "nnUnetActivationStyle" -- U-net activation explained in https://pubmed.ncbi.nlm.nih.gov/33288961/
            * "initialConvolutionalKernelSize[X]" -- Set the first two convolutional layer kernel sizes to X.

    Returns
    -------
    Keras model
        A 2-D keras model defining the U-net network.

    Example
    -------
    >>> model = create_unet_model_2d((128, 128, 1))
    >>> model.summary()
    """

    def nn_unet_activation(x):
        x = InstanceNormalization()(x)
        x = LeakyReLU(alpha=0.01)(x)
        return x

    def attention_gate_2d(x, g, inter_shape):
        x_theta = Conv2D(filters=inter_shape,
                         kernel_size=(1, 1),
                         strides=(1, 1))(x)
        g_phi = Conv2D(filters=inter_shape,
                       kernel_size=(1, 1),
                       strides=(1, 1))(g)
        f = Add()([x_theta, g_phi])
        f = ReLU()(f)
        f_psi = Conv2D(filters=1,
                       kernel_size=(1, 1),
                       strides=(1, 1))(f)
        alpha = Activation('sigmoid')(f_psi)
        attention = multiply([x, alpha])
        return attention

    # Handle additional specific configurations

    initial_convolution_kernel_size = convolution_kernel_size
    add_attention_gating = False
    nn_unet_activation_style = False

    if additional_options is not None:

        if "attentionGating" in additional_options:
            add_attention_gating = True

        if "nnUnetActivationStyle" in additional_options:
            nn_unet_activation_style = True

        option = [o for o in additional_options if o.startswith('initialConvolutionKernelSize')]
        if not not option:
            initial_convolution_kernel_size = option[0].replace("initialConvolutionKernelSize", "")
            initial_convolution_kernel_size = initial_convolution_kernel_size.replace("[", "")
            initial_convolution_kernel_size = int(initial_convolution_kernel_size.replace("]", ""))

    # Specify the number of filters

    if number_of_filters is not None:
        number_of_layers = len(number_of_filters)
    else:
        number_of_filters = list()
        for i in range(number_of_layers):
            number_of_filters.append(number_of_filters_at_base_layer * 2**i)

    inputs = Input(shape = input_image_size)

    # Encoding path

    encoding_convolution_layers = []
    pool = None
    for i in range(number_of_layers):

        if i == 0:
            conv = Conv2D(filters=number_of_filters[i],
                          kernel_size=initial_convolution_kernel_size,
                          padding='same',
                          kernel_regularizer=regularizers.l2(weight_decay))(inputs)
        else:
            conv = Conv2D(filters=number_of_filters[i],
                          kernel_size=convolution_kernel_size,
                          padding='same',
                          kernel_regularizer=regularizers.l2(weight_decay))(pool)

        if nn_unet_activation_style == True:
            conv = nn_unet_activation(conv)
        else:
            conv = ReLU()(conv)

        if dropout_rate > 0.0:
            conv = Dropout(rate=dropout_rate)(conv)

        if i == 0:
            conv = Conv2D(filters=number_of_filters[i],
                          kernel_size=initial_convolution_kernel_size,
                          padding='same')(conv)
        else:
            conv = Conv2D(filters=number_of_filters[i],
                          kernel_size=convolution_kernel_size,
                          padding='same')(conv)

        if nn_unet_activation_style == True:
            conv = nn_unet_activation(conv)
        else:
            conv = ReLU()(conv)

        encoding_convolution_layers.append(conv)

        if i < number_of_layers - 1:
            pool = MaxPooling2D(pool_size=pool_size)(encoding_convolution_layers[i])

    scalar_output = None
    if scalar_output_size > 0:
        scalar_layers = list()
        for i in range(number_of_layers):
            scalar_layers.append(GlobalAveragePooling2D()(encoding_convolution_layers[i]))
        scalar_output = Concatenate()(scalar_layers)
        scalar_output = Dense(units=scalar_output_size,
                              activation=scalar_output_activation)(scalar_output)

    # Decoding path

    outputs = encoding_convolution_layers[number_of_layers - 1]
    for i in range(1, number_of_layers):
        deconv = Conv2DTranspose(filters=number_of_filters[number_of_layers-i-1],
                                 kernel_size=deconvolution_kernel_size,
                                 padding='same',
                                 kernel_regularizer=regularizers.l2(weight_decay))(outputs)
        if nn_unet_activation_style == True:
            deconv = nn_unet_activation(deconv)
        deconv = UpSampling2D(size=pool_size)(deconv)

        if add_attention_gating == True:
            outputs = attention_gate_2d(deconv,
              encoding_convolution_layers[number_of_layers-i-1],
              number_of_filters[number_of_layers-i-1] // 4)
            outputs = Concatenate(axis=3)([deconv, outputs])
        else:
            outputs = Concatenate(axis=3)([deconv, encoding_convolution_layers[number_of_layers-i-1]])

        outputs = Conv2D(filters=number_of_filters[number_of_layers-i-1],
                         kernel_size=convolution_kernel_size,
                         padding='same',
                         kernel_regularizer=regularizers.l2(weight_decay))(outputs)
        if nn_unet_activation_style == True:
            outputs = nn_unet_activation(outputs)
        else:
            outputs = ReLU()(outputs)

        if dropout_rate > 0.0:
            outputs = Dropout(rate=dropout_rate)(outputs)

        outputs = Conv2D(filters=number_of_filters[number_of_layers-i-1],
                         kernel_size=convolution_kernel_size,
                         padding='same',
                         kernel_regularizer=regularizers.l2(weight_decay))(outputs)

        if nn_unet_activation_style == True:
            outputs = nn_unet_activation(outputs)
        else:
            outputs = ReLU()(outputs)

    conv_activation = ''

    if mode == 'sigmoid':
        conv_activation = 'sigmoid'
    elif mode == 'classification':
        conv_activation = 'softmax'
    elif mode == 'regression':
        conv_activation = 'linear'
    else:
        raise ValueError('mode must be either `classification`, `regression` or `sigmoid`.')

    outputs = Conv2D(filters=number_of_outputs,
                     kernel_size=(1, 1),
                     activation = conv_activation,
                     kernel_regularizer=regularizers.l2(weight_decay))(outputs)

    unet_model = None
    if scalar_output_size > 0:
        unet_model = Model(inputs=inputs, outputs=[outputs, scalar_output])
    else:
        unet_model = Model(inputs=inputs, outputs=outputs)

    return unet_model


def create_unet_model_3d(input_image_size,
                         number_of_outputs=2,
                         scalar_output_size = 0,
                         scalar_output_activation = "relu",
                         number_of_layers=4,
                         number_of_filters_at_base_layer=32,
                         number_of_filters=None,
                         convolution_kernel_size=(3, 3, 3),
                         deconvolution_kernel_size=(2, 2, 2),
                         pool_size=(2, 2, 2),
                         strides=(2, 2, 2),
                         dropout_rate=0.0,
                         weight_decay=0.0,
                         mode='classification',
                         additional_options=None
                        ):
    """
    3-D implementation of the U-net deep learning architecture.

    Creates a keras model of the U-net deep learning architecture for image
    segmentation and regression.  More information is provided at the authors'
    website:

            https://lmb.informatik.uni-freiburg.de/people/ronneber/u-net/

    with the paper available here:

            https://arxiv.org/abs/1505.04597

    This particular implementation was influenced by the following python
    implementation:

            https://github.com/joelthelion/ultrasound-nerve-segmentation


    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    number_of_outputs : integer
        Meaning depends on the mode.  For `classification` this is the number of
        segmentation labels.  For `regression` this is the number of outputs.

    scalar_output_size : integer
        If greater than 0, a global average pooling from each
        encoding layer is concatenated to a dense layer as a secondary output.

    scalar_output_activation : string
        Activation for nonzero output scalar.

    number_of_layers : integer
        number of encoding/decoding layers.

    number_of_filters_at_base_layer : integer
        number of filters at the beginning and end of the `U`.  Doubles at each
        descending/ascending layer.

    number_of_filters : tuple
        tuple explicitly setting the number of filters at each layer.  One can
        either set this or number_of_layers and  number_of_filters_at_base_layer.
        Default = None.

    convolution_kernel_size : tuple of length 3
        Defines the kernel size during the encoding.

    deconvolution_kernel_size : tuple of length 3
        Defines the kernel size during the decoding.

    pool_size : tuple of length 3
        Defines the region for each pooling layer.

    strides : tuple of length 3
        Strides for the convolutional layers.

    dropout_rate : scalar
        Float between 0 and 1 to use between dense layers.

    weight_decay :  scalar
        Weighting parameter for L2 regularization of the kernel weights of the
        convolution layers.  Default = 0.0.

    mode :  string
        `classification` `regression`, or `sigmoid`.  Default = `classification`.

    additional_options : string or tuple of strings
        specific configuration add-ons/tweaks:
            * "attentionGating" -- attention-unet variant in https://pubmed.ncbi.nlm.nih.gov/33288961/
            * "nnUnetActivationStyle" -- U-net activation explained in https://pubmed.ncbi.nlm.nih.gov/33288961/
            * "initialConvolutionalKernelSize[X]" -- Set the first two convolutional layer kernel sizes to X.

    Returns
    -------
    Keras model
        A 3-D keras model defining the U-net network.

    Example
    -------
    >>> model = create_unet_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    def nn_unet_activation(x):
        x = InstanceNormalization()(x)
        x = LeakyReLU(alpha=0.01)(x)
        return x

    def attention_gate_3d(x, g, inter_shape):
        x_theta = Conv3D(filters=inter_shape,
                         kernel_size=(1, 1, 1),
                         strides=(1, 1, 1))(x)
        g_phi = Conv3D(filters=inter_shape,
                       kernel_size=(1, 1, 1),
                       strides=(1, 1, 1))(g)
        f = Add()([x_theta, g_phi])
        f = ReLU()(f)
        f_psi = Conv3D(filters=1,
                       kernel_size=(1, 1, 1),
                       strides=(1, 1, 1))(f)
        alpha = Activation('sigmoid')(f_psi)
        attention = multiply([x, alpha])
        return attention

    # Handle additional specific configurations

    initial_convolution_kernel_size = convolution_kernel_size
    add_attention_gating = False
    nn_unet_activation_style = False

    if additional_options is not None:

        if "attentionGating" in additional_options:
            add_attention_gating = True

        if "nnUnetActivationStyle" in additional_options:
            nn_unet_activation_style = True

        option = [o for o in additional_options if o.startswith('initialConvolutionKernelSize')]
        if not not option:
            initial_convolution_kernel_size = option[0].replace("initialConvolutionKernelSize", "")
            initial_convolution_kernel_size = initial_convolution_kernel_size.replace("[", "")
            initial_convolution_kernel_size = int(initial_convolution_kernel_size.replace("]", ""))

    # Specify the number of filters

    if number_of_filters is not None:
        number_of_layers = len(number_of_filters)
    else:
        number_of_filters = list()
        for i in range(number_of_layers):
            number_of_filters.append(number_of_filters_at_base_layer * 2**i)

    inputs = Input(shape = input_image_size)

    # Encoding path

    encoding_convolution_layers = []
    pool = None
    for i in range(number_of_layers):

        if i == 0:
            conv = Conv3D(filters=number_of_filters[i],
                          kernel_size=initial_convolution_kernel_size,
                          padding='same',
                          kernel_regularizer=regularizers.l2(weight_decay))(inputs)
        else:
            conv = Conv3D(filters=number_of_filters[i],
                          kernel_size=convolution_kernel_size,
                          padding='same',
                          kernel_regularizer=regularizers.l2(weight_decay))(pool)

        if nn_unet_activation_style == True:
            conv = nn_unet_activation(conv)
        else:
            conv = ReLU()(conv)

        if dropout_rate > 0.0:
            conv = Dropout(rate=dropout_rate)(conv)

        if i == 0:
            conv = Conv3D(filters=number_of_filters[i],
                          kernel_size=initial_convolution_kernel_size,
                          padding='same')(conv)
        else:
            conv = Conv3D(filters=number_of_filters[i],
                          kernel_size=convolution_kernel_size,
                          padding='same')(conv)

        if nn_unet_activation_style == True:
            conv = nn_unet_activation(conv)
        else:
            conv = ReLU()(conv)

        encoding_convolution_layers.append(conv)

        if i < number_of_layers - 1:
            pool = MaxPooling3D(pool_size=pool_size)(encoding_convolution_layers[i])

    scalar_output = None
    if scalar_output_size > 0:
        scalar_layers = list()
        for i in range(number_of_layers):
            scalar_layers.append(GlobalAveragePooling3D()(encoding_convolution_layers[i]))
        scalar_output = Concatenate()(scalar_layers)
        scalar_output = Dense(units=scalar_output_size,
                              activation=scalar_output_activation)(scalar_output)

    # Decoding path

    outputs = encoding_convolution_layers[number_of_layers - 1]
    for i in range(1, number_of_layers):
        deconv = Conv3DTranspose(filters=number_of_filters[number_of_layers-i-1],
                                 kernel_size=deconvolution_kernel_size,
                                 padding='same',
                                 kernel_regularizer=regularizers.l2(weight_decay))(outputs)
        if nn_unet_activation_style == True:
            deconv = nn_unet_activation(deconv)
        deconv = UpSampling3D(size=pool_size)(deconv)

        if add_attention_gating == True:
            outputs = attention_gate_3d(deconv,
              encoding_convolution_layers[number_of_layers-i-1],
              number_of_filters[number_of_layers-i-1] // 4)
            outputs = Concatenate(axis=4)([deconv, outputs])
        else:
            outputs = Concatenate(axis=4)([deconv, encoding_convolution_layers[number_of_layers-i-1]])

        outputs = Conv3D(filters=number_of_filters[number_of_layers-i-1],
                         kernel_size=convolution_kernel_size,
                         padding='same',
                         kernel_regularizer=regularizers.l2(weight_decay))(outputs)
        if nn_unet_activation_style == True:
            outputs = nn_unet_activation(outputs)
        else:
            outputs = ReLU()(outputs)

        if dropout_rate > 0.0:
            outputs = Dropout(rate=dropout_rate)(outputs)

        outputs = Conv3D(filters=number_of_filters[number_of_layers-i-1],
                         kernel_size=convolution_kernel_size,
                         padding='same',
                         kernel_regularizer=regularizers.l2(weight_decay))(outputs)

        if nn_unet_activation_style == True:
            outputs = nn_unet_activation(outputs)
        else:
            outputs = ReLU()(outputs)

    conv_activation = ''
    if mode == 'sigmoid':
        conv_activation = 'sigmoid'
    elif mode == 'classification':
        conv_activation = 'softmax'
    elif mode == 'regression':
        conv_activation = 'linear'
    else:
        raise ValueError('mode must be either `classification`, `regression` or `sigmoid`.')

    outputs = Conv3D(filters=number_of_outputs,
                     kernel_size=(1, 1, 1),
                     activation=conv_activation,
                     kernel_regularizer=regularizers.l2(weight_decay))(outputs)

    unet_model = None
    if scalar_output_size > 0:
        unet_model = Model(inputs=inputs, outputs=[outputs, scalar_output])
    else:
        unet_model = Model(inputs=inputs, outputs=outputs)

    return unet_model
